"""
Lunch Money Python SDK
"""

from ._version import __application__, __author__, __email__, __version__
from .exceptions import LunchMoneyError
from .models._lunchmoney import LunchMoney
from .models.transactions import (
    TransactionInsertObject,
    TransactionSplitObject,
    TransactionUpdateObject,
)

__all__ = [
    "LunchMoney",
    "LunchMoneyError",
    "TransactionInsertObject",
    "TransactionUpdateObject",
    "TransactionSplitObject",
    "__application__",
    "__version__",
    "__author__",
    "__email__",
]



"""
Lunchable Plugin for Splitwise
"""

import datetime
import logging
from math import floor
from os import getenv
from random import shuffle
from textwrap import dedent
from typing import Any, Dict, List, Optional, Tuple, Union

from lunchable import LunchMoney, __application__
from lunchable.exceptions import LunchMoneyImportError
from lunchable.models import (
    AssetsObject,
    CategoriesObject,
    TagsObject,
    TransactionInsertObject,
    TransactionObject,
    TransactionSplitObject,
    TransactionUpdateObject,
)
from lunchable.plugins.splitlunch._config import SplitLunchConfig
from lunchable.plugins.splitlunch.exceptions import SplitLunchError
from lunchable.plugins.splitlunch.models import SplitLunchExpense

logger = logging.getLogger(__name__)

try:
    import splitwise  # type: ignore
    from dateutil.tz import tzlocal
except ImportError as ie:
    logger.exception(ie)
    _pip_extra_error = (
        "Looks like you don't have the Splitwise plugin installed: "
        f"`pip install {__application__}[splitlunch]`"
    )
    raise LunchMoneyImportError(_pip_extra_error)


class SplitLunch(splitwise.Splitwise):
    """
    Lunchable Plugin For Interacting With Splitwise

    This plugin supports different operations, and some of those operations
    have prerequisites:

    1) It supports the auto-importing of Splitwise expenses into Lunch Money
    transactions. This requires a manual asset exist in your Lunch Money
    account with "Splitwise" in the Name.

        Prerequisites:
           - Accounts:
                - Splitwise must be in the account name

    2) It supports the creation of Splitwise transactions directly from synced Lunch Money
    accounts. This syncing requires you create a tag called `SplitLunchImport`. Transactions
    with this tag will be created in Splitwise with your "financial partner". Once transactions
    are created in Splitwise they will be split in half in Lunch Money. Half of the split will be
    marked in the `Reimbursement` category which must be created.

        Prerequisites:
           - Financial Partners:
                - If you only have one friend in Splitwise, this is your Financial Partner
                - Financial Partners must be specified by their Splitwise ID or Email Address
           - Tags:
                - `SplitLunchImport`
           - Categories:
                - `Reimbursement`

    3) It supports a workflow where you mark transactions as split (identical to scenario #2)
    without importing them into Splitwise. This syncing requires you create a tag
    called `SplitLunch` and a category named `Reimbursement`

        Prerequisites:
           - Tags:
                - `SplitLunch`
           - Categories:
                - `Reimbursement`

    4) It supports a workflow where you mark transactions as `Reimbursed` and import them to
    Splitwise with the total completely owed by your financial partner.

        Prerequisites:
           - Financial Partners:
                - If you only have one friend in Splitwise, this is your Financial Partner
                - Financial Partners must be specified by their Splitwise ID or Email Address
           - Tags:
                - `SplitLunchDirectImport`
           - Categories:
                - `Reimbursement`

    Parameters
    ----------
    financial_partner_id: Optional[int]
        Splitwise User ID of financial partner
    financial_partner_email: Optional[str]
        Splitwise linked email address of financial partner
    consumer_key: Optional[str]
        Consumer Key provided by Splitwise. Defaults to `SPLITWISE_CONSUMER_KEY` environment
        variable
    consumer_secret: Optional[str]
        Consumer Key provided by Splitwise. Defaults to `SPLITWISE_CONSUMER_SECRET`
        environment variable
    api_key: Optional[str]
        Consumer Key provided by Splitwise. Defaults to `SPLITWISE_API_KEY` environment
        variable.
    lunchable_client: LunchMoney
        Instantiated LunchMoney object to use as internal client. One will
        be created using environment variables otherwise.
    """

    def __init__(
        self,
        lunch_money_access_token: Optional[str] = None,
        financial_partner_id: Optional[int] = None,
        financial_partner_email: Optional[str] = None,
        consumer_key: Optional[str] = None,
        consumer_secret: Optional[str] = None,
        api_key: Optional[str] = None,
        lunchable_client: Optional[LunchMoney] = None,
    ):
        """
        Initialize the Parent Class with some additional properties

        Parameters
        ----------
        financial_partner_id: Optional[int]
            Splitwise User ID of financial partner
        financial_partner_email: Optional[str]
            Splitwise linked email address of financial partner
        consumer_key: Optional[str]
            Consumer Key provided by Splitwise. Defaults to `SPLITWISE_CONSUMER_KEY` environment
            variable
        consumer_secret: Optional[str]
            Consumer Key provided by Splitwise. Defaults to `SPLITWISE_CONSUMER_SECRET`
            environment variable
        api_key: Optional[str]
            Consumer Key provided by Splitwise. Defaults to `SPLITWISE_API_KEY` environment
            variable.
        lunch_money_access_token: Optional[str]
            Lunch Money Access Token. Will be inherited from `LUNCHMONEY_ACCESS_TOKEN`
            environment variable if not provided.
        lunchable_client: LunchMoney
            Instantiated LunchMoney object to use as internal client. One will
            be created using environment variables otherwise.
        """
        init_kwargs = self._get_splitwise_init_kwargs(
            consumer_key=consumer_key, consumer_secret=consumer_secret, api_key=api_key
        )
        super(SplitLunch, self).__init__(**init_kwargs)
        self.current_user: splitwise.CurrentUser = self.getCurrentUser()
        self.financial_partner: splitwise.Friend = self.get_friend(
            friend_id=financial_partner_id, email_address=financial_partner_email
        )
        self.last_check: Optional[datetime.datetime] = None
        self.lunchable = (
            LunchMoney(access_token=lunch_money_access_token)
            if lunchable_client is None
            else lunchable_client
        )
        self._none_tag = TagsObject(id=0, name="SplitLunchPlaceholder")
        self.splitwise_tag = self._none_tag.copy()
        self.splitlunch_tag = self._none_tag.copy()
        self.splitlunch_import_tag = self._none_tag.copy()
        self.splitlunch_direct_import_tag = self._none_tag.copy()
        self._get_splitwise_tags()
        self.earliest_start_date = datetime.date(1812, 1, 1)
        today = datetime.date.today()
        self.latest_end_date = datetime.date(today.year + 10, 12, 31)
        self.splitwise_asset = self._get_splitwise_asset()
        self.reimbursement_category = self._get_reimbursement_category()

    def __repr__(self):
        """
        String Representation

        Returns
        -------
        str
        """
        return f"<Splitwise: {self.current_user.email}>"

    @classmethod
    def _split_amount(cls, amount: float, splits: int) -> Tuple[float, ...]:
        """
        Split a money amount into fair shares

        Parameters
        ----------
        amount: float
        splits: int

        Returns
        -------
        Tuple[float]
        """
        try:
            assert amount == round(amount, 2)
        except AssertionError:
            raise SplitLunchError(
                f"{amount} caused an error, you must provide a real " "spending amount."
            )
        equal_shares = round(amount, 2) / splits
        remainder_dollars = floor(equal_shares)
        remainder_cents = floor((equal_shares - remainder_dollars) * 100) / 100
        remainder_left = round(
            (equal_shares - remainder_dollars - remainder_cents) * splits * 100, 0
        )
        owed_amount = remainder_dollars + remainder_cents
        return_amounts = [owed_amount for _ in range(splits)]
        for i in range(int(remainder_left)):
            return_amounts[i] += 0.010
        shuffle(return_amounts)
        return tuple([round(item, 2) for item in return_amounts])

    @classmethod
    def split_a_transaction(cls, amount: Union[float, int]) -> Tuple[float, ...]:
        """
        Split a Transaction into Two

        Split a bill into a tuple of two amounts (and take care
        of the extra penny if needed)

        Parameters
        ----------
        amount: A Currency amount (no more precise than cents)

        Returns
        -------
        tuple
            A tuple is returned with each participant's amount
        """
        amounts_due = cls._split_amount(amount=amount, splits=2)
        return amounts_due

    def create_self_paid_expense(
        self, amount: float, description: str
    ) -> SplitLunchExpense:
        """
        Create and Submit a Splitwise Expense

        Parameters
        ----------
        amount: float
            Transaction Amount
        description: str
            Transaction Description

        Returns
        -------
        Expense
        """
        # CREATE THE NEW EXPENSE OBJECT
        new_expense = splitwise.Expense()
        new_expense.setDescription(desc=description)
        # GET AND SET AMOUNTS OWED
        primary_user_owes, financial_partner_owes = self.split_a_transaction(
            amount=amount
        )
        new_expense.setCost(cost=amount)
        # CONFIGURE PRIMARY USER
        primary_user = splitwise.user.ExpenseUser()
        primary_user.setId(id=self.current_user.id)
        primary_user.setPaidShare(paid_share=amount)
        primary_user.setOwedShare(owed_share=primary_user_owes)
        # CONFIGURE SECONDARY USER
        financial_partner = splitwise.user.ExpenseUser()
        financial_partner.setId(id=self.financial_partner.id)
        financial_partner.setPaidShare(paid_share=0.00)
        financial_partner.setOwedShare(owed_share=financial_partner_owes)
        # ADD USERS AND REPAYMENTS TO EXPENSE
        new_expense.addUser(user=primary_user)
        new_expense.addUser(user=financial_partner)
        # SUBMIT THE EXPENSE AND GET THE RESPONSE
        expense_response: splitwise.Expense
        expense_response, expense_errors = self.createExpense(expense=new_expense)
        try:
            assert expense_errors is None
        except AssertionError:
            raise SplitLunchError(expense_errors["base"][0])
        logger.info("Expense Created: %s", expense_response.id)
        message = f"Created via SplitLunch: {datetime.datetime.now()}"
        self.createComment(expense_id=expense_response.id, content=message)
        pydantic_response = self.splitwise_to_pydantic(expense=expense_response)
        return pydantic_response

    def create_expense_on_behalf_of_partner(
        self, amount: float, description: str
    ) -> SplitLunchExpense:
        """
        Create and Submit a Splitwise Expense on behalf of your financial partner.

        This expense will be completely owed by the partner and maked as reimbursed.

        Parameters
        ----------
        amount: float
            Transaction Amount
        description: str
            Transaction Description

        Returns
        -------
        Expense
        """
        # CREATE THE NEW EXPENSE OBJECT
        new_expense = splitwise.Expense()
        new_expense.setDescription(desc=description)
        # GET AND SET AMOUNTS OWED
        new_expense.setCost(cost=amount)
        # CONFIGURE PRIMARY USER
        primary_user = splitwise.user.ExpenseUser()
        primary_user.setId(id=self.current_user.id)
        primary_user.setPaidShare(paid_share=amount)
        primary_user.setOwedShare(owed_share=0.00)
        # CONFIGURE SECONDARY USER
        financial_partner = splitwise.user.ExpenseUser()
        financial_partner.setId(id=self.financial_partner.id)
        financial_partner.setPaidShare(paid_share=0.00)
        financial_partner.setOwedShare(owed_share=amount)
        # ADD USERS AND REPAYMENTS TO EXPENSE
        new_expense.addUser(user=primary_user)
        new_expense.addUser(user=financial_partner)
        # SUBMIT THE EXPENSE AND GET THE RESPONSE
        expense_response: splitwise.Expense
        expense_response, expense_errors = self.createExpense(expense=new_expense)
        try:
            assert expense_errors is None
        except AssertionError:
            raise SplitLunchError(expense_errors["base"][0])
        logger.info("Expense Created: %s", expense_response.id)
        message = f"Created via SplitLunch: {datetime.datetime.now()}"
        self.createComment(expense_id=expense_response.id, content=message)
        pydantic_response = self.splitwise_to_pydantic(expense=expense_response)
        return pydantic_response

    def get_friend(
        self, email_address: Optional[str] = None, friend_id: Optional[int] = None
    ) -> Optional[splitwise.Friend]:
        """
        Retrieve a Financial Partner by Email Address

        Parameters
        ----------
        email_address: str
            Email Address of Friend's user in Splitwise
        friend_id: Optional[int]
            Splitwise friend ID. Notice the friend ID in the following
            URL: https://secure.splitwise.com/#/friends/12345678

        Returns
        -------
        Optional[splitwise.Friend]
        """
        friend_list: List[splitwise.Friend] = self.getFriends()
        if len(friend_list) == 1:
            return friend_list[0]
        for friend in friend_list:
            if friend_id is not None and friend.id == friend_id:
                return friend
            elif (
                email_address is not None
                and friend.email.lower() == email_address.lower()
            ):
                return friend
        return None

    def get_expenses(
        self,
        offset: Optional[int] = None,
        limit: Optional[int] = None,
        group_id: Optional[int] = None,
        friendship_id: Optional[int] = None,
        dated_after: Optional[datetime.datetime] = None,
        dated_before: Optional[datetime.datetime] = None,
        updated_after: Optional[datetime.datetime] = None,
        updated_before: Optional[datetime.datetime] = None,
    ) -> List[SplitLunchExpense]:
        """
        Get Splitwise Expenses

        Parameters
        ----------
        offset: Optional[int]
            Number of expenses to be skipped
        limit: Optional[int]
            Number of expenses to be returned
        group_id: Optional[int]
            GroupID of the expenses
        friendship_id: Optional[int]
            FriendshipID of the expenses
        dated_after: Optional[datetime.datetime]
            ISO 8601 Date time. Return expenses later that this date
        dated_before: Optional[datetime.datetime]
            ISO 8601 Date time. Return expenses earlier than this date
        updated_after: Optional[datetime.datetime]
            ISO 8601 Date time. Return expenses updated after this date
        updated_before: Optional[datetime.datetime]
            ISO 8601 Date time. Return expenses updated before this date

        Returns
        -------
        List[SplitLunchExpense]
        """
        expenses = self.getExpenses(
            offset=offset,
            limit=limit,
            group_id=group_id,
            friendship_id=friendship_id,
            dated_after=dated_after,
            dated_before=dated_before,
            updated_after=updated_after,
            updated_before=updated_before,
        )
        pydantic_expenses = [
            self.splitwise_to_pydantic(expense) for expense in expenses
        ]
        return pydantic_expenses

    @classmethod
    def _get_splitwise_init_kwargs(
        cls,
        consumer_key: Optional[str] = None,
        consumer_secret: Optional[str] = None,
        api_key: Optional[str] = None,
    ) -> Dict[str, Any]:
        """
        Get the Splitwise Kwargs

        Parameters
        ----------
        consumer_key: Optional[str]
        consumer_secret: Optional[str]
        api_key: Optional[str]
        """
        if consumer_key is None:
            consumer_key = getenv("SPLITWISE_CONSUMER_KEY")
        if consumer_secret is None:
            consumer_secret = getenv("SPLITWISE_CONSUMER_SECRET")
        if api_key is None:
            api_key = getenv("SPLITWISE_API_KEY", None)
        init_kwargs = dict(
            consumer_key=consumer_key, consumer_secret=consumer_secret, api_key=api_key
        )
        if consumer_key is None or consumer_secret is None or api_key is None:
            error_message = (
                dedent(
                    """
            You must set your Splitwise credentials explicitly or by assigning
            the `SPLITWISE_CONSUMER_KEY`, `SPLITWISE_CONSUMER_SECRET`, and the
            `SPLITWISE_API_KEY`environment variables
            """
                )
                .replace("\n", " ")
                .replace("  ", " ")
            )
            logger.error(error_message)
            raise SplitLunchError(error_message)
        return init_kwargs

    def splitwise_to_pydantic(self, expense: splitwise.Expense) -> SplitLunchExpense:
        """
        Convert Splitwise Object to Pydantic

        Parameters
        ----------
        expense: splitwise.Expense

        Returns
        -------
        SplitLunchExpense
        """
        financial_impact, self_paid = self._get_splitwise_impact(expense=expense)
        if expense.payment is True and financial_impact < 0:
            personal_share = abs(financial_impact)
        elif expense.payment is True and financial_impact > 0:
            personal_share = 0
        elif self_paid is True:
            personal_share = round(float(expense.cost) - financial_impact, 2)
        else:
            personal_share = abs(financial_impact)
        expense = SplitLunchExpense(
            splitwise_id=expense.id,
            original_amount=expense.cost,
            financial_impact=financial_impact,
            personal_share=personal_share,
            self_paid=self_paid,
            description=expense.description,
            category=expense.category.name,
            details=expense.details,
            payment=expense.payment,
            date=expense.date,
            users=[user.id for user in expense.users],
            created_at=expense.created_at,
            updated_at=expense.updated_at,
            deleted_at=expense.deleted_at,
            deleted=True if expense.deleted_at is not None else False,
        )
        return expense

    def _get_expense_impact(self, expense: splitwise.Expense) -> Tuple[float, bool]:
        """
        Get the Financial Impact of a Splitwise Expense

        Parameters
        ----------
        expense: splitwise.Expense

        Returns
        -------
        Tuple[float, bool]
        """
        financial_impact = 0.00
        self_paid = True
        if len(expense.repayments) >= 1:
            for debt in expense.repayments:
                if debt.fromUser == self.current_user.id:
                    self_paid = False
                    financial_impact -= float(debt.amount)
                elif debt.toUser == self.current_user.id:
                    financial_impact += float(debt.amount)
        elif len(expense.repayments) == 0:
            assert len(expense.users) == 1
            assert expense.users[0].id == self.current_user.id
        return financial_impact, self_paid

    def _get_payment_impact(self, expense: splitwise.Expense) -> Tuple[float, bool]:
        """
        Get the Financial Impact of a Splitwise Payment

        Parameters
        ----------
        expense: splitwise.Expense

        Returns
        -------
        Tuple[float, bool]
        """
        financial_impact = 0.00
        self_paid = True
        if len(expense.repayments) >= 1:
            for debt in expense.repayments:
                if debt.fromUser == self.current_user.id:
                    self_paid = False
                    financial_impact += float(debt.amount)
                elif debt.toUser == self.current_user.id:
                    financial_impact -= float(debt.amount)
        elif len(expense.repayments) == 0:
            assert expense.users[0].id == self.current_user.id
            financial_impact -= float(expense.cost)
        return financial_impact, self_paid

    def _get_splitwise_asset(self) -> Optional[AssetsObject]:
        """
        Get the Splitwise asset

        Parse a user's Lunch Money accounts and return the manually managed
        Splitwise account asset object

        Returns
        -------
        AssetsObject
        """
        assets = self.lunchable.get_assets()
        splitwise_assets = list()
        for asset in assets:
            if (
                asset.institution_name is not None
                and "splitwise" in asset.institution_name.lower()
            ):
                splitwise_assets.append(asset)
        if len(splitwise_assets) == 0:
            return None
        elif len(splitwise_assets) > 1:
            raise SplitLunchError(
                "SplitLunch requires an manually managed Splitwise asset. "
                "Make sure you have a single account where 'Splitwise' "
                "is in the asset's `Institution Name`."
            )
        else:
            return splitwise_assets[0]

    def _get_reimbursement_category(self) -> Optional[CategoriesObject]:
        """
        Get the Reimbusement Category

        Parse a user's Lunch Money categories and return the Reimbursement
        category

        Returns
        -------
        CategoriesObject
        """
        categories = self.lunchable.get_categories()
        reimbursement_list = list()
        for category in categories:
            if "reimbursement" == category.name.strip().lower():
                reimbursement_list.append(category)
        if len(reimbursement_list) != 1:
            return None
        return reimbursement_list[0]

    def _get_splitwise_impact(self, expense: splitwise.Expense) -> Tuple[float, bool]:
        """
        Get the Financial Impact of a Splitwise Transaction

        Parameters
        ----------
        expense: splitwise.Expense

        Returns
        -------
        Tuple[float, bool]
        """
        if expense.payment is True:
            financial_impact, self_paid = self._get_payment_impact(expense=expense)
        else:
            financial_impact, self_paid = self._get_expense_impact(expense=expense)
        return financial_impact, self_paid

    def _get_splitwise_tags(self) -> None:
        """
        Get Lunch Money Tags to Interact with

        Returns
        -------
        Dict[str, int]
        """
        all_tags = self.lunchable.get_tags()
        for tag in all_tags:
            if tag.name.lower() == SplitLunchConfig.splitlunch_tag.lower():
                self.splitlunch_tag = tag
            elif tag.name.lower() == SplitLunchConfig.splitwise_tag.lower():
                self.splitwise_tag = tag
            elif tag.name.lower() == SplitLunchConfig.splitlunch_import_tag.lower():
                self.splitlunch_import_tag = tag
            elif (
                tag.name.lower()
                == SplitLunchConfig.splitlunch_direct_import_tag.lower()
            ):
                self.splitlunch_direct_import_tag = tag

    def _raise_nonexistent_tag_error(self, tags: List[str]) -> None:
        """
        Raise a warning for specific SplitLunch Tags

        tags: List[str]
            A list of tags to raise the error for
        """
        if (
            self.splitlunch_tag == self._none_tag
            and SplitLunchConfig.splitlunch_tag in tags
        ):
            error_message = (
                f"a `{SplitLunchConfig.splitlunch_tag}` tag is required. "
                f"This tag is used for splitting transactions in half and have half "
                f"marked as reimbursed."
            )
            raise SplitLunchError(error_message)
        if (
            self.splitwise_tag == self._none_tag
            and SplitLunchConfig.splitwise_tag in tags
        ):
            error_message = (
                f"a `{SplitLunchConfig.splitwise_tag}` tag is required. "
                f"This tag is used for splitting transactions in half and have half "
                f"marked as reimbursed."
            )
            raise SplitLunchError(error_message)
        if (
            self.splitlunch_import_tag == self._none_tag
            and SplitLunchConfig.splitlunch_import_tag in tags
        ):
            error_message = (
                f"a `{SplitLunchConfig.splitlunch_import_tag}` tag is required. "
                f"This tag is used for creating Splitwise transactions directly from "
                f"Lunch Money transactions. These transactions will be split in half,"
                f"and have one half marked as reimbursed."
            )
            raise SplitLunchError(error_message)
        if (
            self.splitlunch_direct_import_tag == self._none_tag
            and SplitLunchConfig.splitlunch_direct_import_tag in tags
        ):
            error_message = (
                f"a `{SplitLunchConfig.splitlunch_direct_import_tag}` tag is "
                "required. This tag is used for creating Splitwise transactions "
                "directly from Lunch Money transactions. These transactions will "
                "be completely owed by your financial partner."
            )
            raise SplitLunchError(error_message)

    def get_splitlunch_tagged_transactions(
        self,
        start_date: Optional[datetime.date] = None,
        end_date: Optional[datetime.date] = None,
    ) -> List[TransactionObject]:
        """
        Retrieve all transactions with the "Splitlunch" Tag

        Parameters
        ----------
        start_date: Optional[datetime.date]
        end_date : Optional[datetime.date]

        Returns
        -------
        List[TransactionObject]
        """
        if start_date is None:
            start_date = self.earliest_start_date
        if end_date is None:
            end_date = self.latest_end_date
        self._raise_nonexistent_tag_error(tags=[SplitLunchConfig.splitlunch_tag])
        transactions = self.lunchable.get_transactions(
            tag_id=self.splitlunch_tag.id, start_date=start_date, end_date=end_date
        )
        return transactions

    def get_splitlunch_import_tagged_transactions(
        self,
        start_date: Optional[datetime.date] = None,
        end_date: Optional[datetime.date] = None,
    ) -> List[TransactionObject]:
        """
        Retrieve all transactions with the "SplitLunchImport" Tag

        Parameters
        ----------
        start_date: Optional[datetime.date]
        end_date : Optional[datetime.date]

        Returns
        -------
        List[TransactionObject]
        """
        if start_date is None:
            start_date = self.earliest_start_date
        if end_date is None:
            end_date = self.latest_end_date
        self._raise_nonexistent_tag_error(tags=[SplitLunchConfig.splitlunch_import_tag])
        transactions = self.lunchable.get_transactions(
            tag_id=self.splitlunch_import_tag.id,
            start_date=start_date,
            end_date=end_date,
        )
        return transactions

    def get_splitlunch_direct_import_tagged_transactions(
        self,
        start_date: Optional[datetime.date] = None,
        end_date: Optional[datetime.date] = None,
    ) -> List[TransactionObject]:
        """
        Retrieve all transactions with the "SplitLunchDirectImport" Tag

        Parameters
        ----------
        start_date: Optional[datetime.date]
        end_date : Optional[datetime.date]

        Returns
        -------
        List[TransactionObject]
        """
        if start_date is None:
            start_date = self.earliest_start_date
        if end_date is None:
            end_date = self.latest_end_date
        self._raise_nonexistent_tag_error(
            tags=[SplitLunchConfig.splitlunch_direct_import_tag]
        )
        transactions = self.lunchable.get_transactions(
            tag_id=self.splitlunch_direct_import_tag.id,
            start_date=start_date,
            end_date=end_date,
        )
        return transactions

    def get_splitwise_tagged_transactions(
        self,
        start_date: Optional[datetime.date] = None,
        end_date: Optional[datetime.date] = None,
    ) -> List[TransactionObject]:
        """
        Retrieve all transactions with the "Splitwise" Tag

        Parameters
        ----------
        start_date: Optional[datetime.date]
        end_date : Optional[datetime.date]

        Returns
        -------
        List[TransactionObject]
        """
        if start_date is None:
            start_date = self.earliest_start_date
        if end_date is None:
            end_date = self.latest_end_date
        self._raise_nonexistent_tag_error(tags=[SplitLunchConfig.splitwise_tag])
        transactions = self.lunchable.get_transactions(
            tag_id=self.splitwise_tag.id, start_date=start_date, end_date=end_date
        )
        return transactions

    def make_splitlunch(self, tag_transactions: bool = False) -> List[Dict[int, Any]]:
        """
        Operate on `SplitLunch` tagged transactions

        Split all transactions with the `SplitLunch` tag in half. One of these
        new splits will be recategorized to `Reimbursement`. Both new splits will receive
        the `Splitwise` tag without any preexisting tags.
        """
        if self.reimbursement_category is None:
            self._raise_category_reimbursement_error()
            raise ValueError("ReimbursementCategory")
        split_transaction_ids = list()
        tagged_objects = self.get_splitlunch_tagged_transactions()
        for transaction in tagged_objects:
            # Split the Original Amount
            amount_1, amount_2 = self.split_a_transaction(amount=transaction.amount)
            # Generate the First Split
            split_object = TransactionSplitObject(
                date=transaction.date,
                category_id=transaction.category_id,
                notes=transaction.notes,
                amount=amount_1,
            )
            # Generate the second split as a copy, change some properties
            reimbursement_object = split_object.copy()
            reimbursement_object.amount = amount_2
            reimbursement_object.category_id = self.reimbursement_category.id
            logger.debug(
                "Splitting transaction: %s -> (%s, %s)",
                transaction.amount,
                amount_1,
                amount_2,
            )

            update_response = self.lunchable.update_transaction(
                transaction_id=transaction.id,
                split=[split_object, reimbursement_object],
            )
            # Tag each of the new transactions generated
            split_transaction_ids.append({transaction.id: update_response["split"]})
            for split_transaction_id in update_response["split"]:
                update_tags = transaction.tags if transaction.tags is not None else []
                tags = [
                    tag.name
                    for tag in update_tags
                    if tag is not None
                    and tag.name.lower() != self.splitlunch_tag.name.lower()
                ]
                if self.splitwise_tag.name not in tags and tag_transactions is True:
                    self._raise_nonexistent_tag_error(
                        tags=[SplitLunchConfig.splitwise_tag]
                    )
                    tags.append(self.splitwise_tag.name)
                tag_update = TransactionUpdateObject(tags=tags)
                self.lunchable.update_transaction(
                    transaction_id=split_transaction_id, transaction=tag_update
                )
        return split_transaction_ids

    def make_splitlunch_import(
        self, tag_transactions: bool = False
    ) -> List[Dict[str, Any]]:
        """
        Operate on `SplitLunchImport` tagged transactions

        Send a transaction to Splitwise and then split the original transaction in Lunch Money.
        One of these new splits will be recategorized to `Reimbursement`. Both new splits
        will receive the `Splitwise` tag without the `SplitLunchImport` tag. Any other tags will be
        reapplied.

        Parameters
        ----------
        tag_transactions : bool
            Whether to tag the transactions with the `Splitwise` tag after splitting them.
            Defaults to False which

        Returns
        -------
        List[Dict[str, Any]]
        """
        self._raise_financial_partner_error()
        if self.reimbursement_category is None:
            self._raise_category_reimbursement_error()
            raise ValueError("ReimbursementCategory")
        tagged_objects = self.get_splitlunch_import_tagged_transactions()
        update_responses = list()
        for transaction in tagged_objects:
            # Split the Original Amount
            description = str(transaction.payee)
            if transaction.notes is not None:
                description = f"{transaction.payee} - {transaction.notes}"
            new_transaction = self.create_self_paid_expense(
                amount=transaction.amount, description=description
            )
            notes = f"Splitwise ID: {new_transaction.splitwise_id}"
            if transaction.notes is not None:
                notes = f"{transaction.notes} || {notes}"
            split_object = TransactionSplitObject(
                date=transaction.date,
                category_id=transaction.category_id,
                notes=notes,
                amount=new_transaction.personal_share,
            )
            reimbursement_object = split_object.copy()
            reimbursement_object.amount = round(
                transaction.amount - new_transaction.personal_share, 2
            )
            reimbursement_object.category_id = self.reimbursement_category.id
            logger.debug(
                f"Transaction split by Splitwise: {transaction.amount} -> "
                f"({split_object.amount}, {reimbursement_object.amount})"
            )
            update_response = self.lunchable.update_transaction(
                transaction_id=transaction.id,
                split=[split_object, reimbursement_object],
            )
            formatted_update_response = dict(
                original_id=transaction.id,
                payee=transaction.payee,
                amount=transaction.amount,
                reimbursement_amount=reimbursement_object.amount,
                notes=transaction.notes,
                splitwise_id=new_transaction.splitwise_id,
                updated=update_response["updated"],
                split=update_response["split"],
            )
            update_responses.append(formatted_update_response)
            # Tag each of the new transactions generated
            for split_transaction_id in update_response["split"]:
                update_tags = transaction.tags or []
                tags = [
                    tag.name
                    for tag in update_tags
                    if tag.name.lower() != self.splitlunch_import_tag.name.lower()
                ]
                if self.splitwise_tag.name not in tags and tag_transactions is True:
                    self._raise_nonexistent_tag_error(
                        tags=[SplitLunchConfig.splitwise_tag]
                    )
                    tags.append(self.splitwise_tag.name)
                tag_update = TransactionUpdateObject(tags=tags)
                self.lunchable.update_transaction(
                    transaction_id=split_transaction_id, transaction=tag_update
                )
        return update_responses

    def make_splitlunch_direct_import(
        self, tag_transactions: bool = False
    ) -> List[Dict[str, Any]]:
        """
        Operate on `SplitLunchDirectImport` tagged transactions

        Send a transaction to Splitwise and then mark the transaction under the
        `Reimbursement` category. The sum of the transaction will be completely owed
        by the financial partner.

        Parameters
        ----------
        tag_transactions : bool
            Whether to tag the transactions with the `Splitwise` tag after splitting them.
            Defaults to False which
        """
        self._raise_financial_partner_error()
        if self.reimbursement_category is None:
            self._raise_category_reimbursement_error()
            raise ValueError("ReimbursementCategory")
        tagged_objects = self.get_splitlunch_direct_import_tagged_transactions()
        update_responses = list()
        for transaction in tagged_objects:
            # Split the Original Amount
            description = str(transaction.payee)
            if transaction.notes is not None:
                description = f"{transaction.payee} - {transaction.notes}"
            new_transaction = self.create_expense_on_behalf_of_partner(
                amount=transaction.amount, description=description
            )
            notes = f"Splitwise ID: {new_transaction.splitwise_id}"
            if transaction.notes is not None:
                notes = f"{transaction.notes} || {notes}"
            existing_tags = transaction.tags or []
            tags = [
                tag.name
                for tag in existing_tags
                if tag.name.lower() != self.splitlunch_direct_import_tag.name.lower()
            ]
            if self.splitwise_tag.name not in tags and tag_transactions is True:
                self._raise_nonexistent_tag_error(tags=[SplitLunchConfig.splitwise_tag])
                tags.append(self.splitwise_tag.name)
            update = TransactionUpdateObject(
                category_id=self.reimbursement_category.id, tags=tags, notes=notes
            )
            response = self.lunchable.update_transaction(
                transaction_id=transaction.id, transaction=update
            )
            formatted_update_response = dict(
                original_id=transaction.id,
                payee=transaction.payee,
                amount=transaction.amount,
                reimbursement_amount=transaction.amount,
                notes=transaction.notes,
                splitwise_id=new_transaction.splitwise_id,
                updated=response["updated"],
                split=None,
            )
            update_responses.append(formatted_update_response)
        return update_responses

    def splitwise_to_lunchmoney(self, expenses: List[SplitLunchExpense]) -> List[int]:
        """
        Ingest Splitwise Expenses into Lunch Money

        This function inserts splitwise expenses into Lunch Money. If an expense
        is not a payment, not deleted, and not self-paid it qualifies for ingestion. Otherwise
        it will be ignored.

        Parameters
        ----------
        expenses: List[SplitLunchExpense]

        Returns
        -------
        List[int]
            New Lunch Money transaction IDs
        """
        if self.splitwise_asset is None:
            self._raise_splitwise_asset_error()
            raise ValueError("SplitwiseAsset")
        batch = []
        new_transaction_ids = []
        filtered_expenses = self.filter_relevant_splitwise_expenses(expenses=expenses)
        for splitwise_transaction in filtered_expenses:
            new_lunchmoney_transaction = TransactionInsertObject(
                date=splitwise_transaction.date.astimezone(tzlocal()),
                payee=splitwise_transaction.description,
                amount=splitwise_transaction.personal_share,
                asset_id=self.splitwise_asset.id,
                external_id=splitwise_transaction.splitwise_id,
            )
            batch.append(new_lunchmoney_transaction)
            if len(batch) == 10:
                new_ids = self.lunchable.insert_transactions(
                    transactions=batch, apply_rules=True
                )
                new_transaction_ids += new_ids
                batch = []
        if len(batch) > 0:
            new_ids = self.lunchable.insert_transactions(
                transactions=batch, apply_rules=True
            )
            new_transaction_ids += new_ids
        return new_transaction_ids

    @staticmethod
    def filter_relevant_splitwise_expenses(
        expenses: List[SplitLunchExpense],
    ) -> List[SplitLunchExpense]:
        """
        Filter Expenses in Splitwise into relevant expenses.

        This filtering action is important to understand when seeing why not
        all transactions from Splitwise end up flowing into Lunch Money.

        1) It filters out deleted expenses

        2) It filters out `self-paid` expenses. A `self-paid` expense is an expense in Splitwise
        where you originated the payment. This is excluded because it is assumed that these
        transactions will have already been imported via a different account.

        3) It filters out payments. Payments are excluded because it is assumed that these
        transactions will have already been imported via a different account.

        Parameters
        ----------
        expenses: List[SplitLunchExpense]

        Returns
        -------
        List[SplitLunchExpense]
        """
        filtered_expenses = list()
        for splitwise_transaction in expenses:
            if all(
                [
                    splitwise_transaction.deleted is False,
                    splitwise_transaction.payment is False,
                    splitwise_transaction.self_paid is False,
                ]
            ):
                filtered_expenses.append(splitwise_transaction)

        return filtered_expenses

    def get_splitwise_balance(self) -> float:
        """
        Get the net balance in Splitwise

        Returns
        -------
        float
        """
        groups = self.getGroups()
        total_balance = 0.00
        for group in groups:
            for debt in group.simplified_debts:
                if debt.fromUser == self.current_user.id:
                    total_balance -= float(debt.amount)
                elif debt.toUser == self.current_user.id:
                    total_balance += float(debt.amount)
        return total_balance

    def update_splitwise_balance(self) -> AssetsObject:
        """
        Get and update the Splitwise Asset in Lunch Money

        Returns
        -------
        AssetsObject
            Updated  balance
        """
        if self.splitwise_asset is None:
            self._raise_splitwise_asset_error()
            raise ValueError("SplitwiseAsset")
        balance = self.get_splitwise_balance()
        if balance != self.splitwise_asset.balance:
            updated_asset = self.lunchable.update_asset(
                asset_id=self.splitwise_asset.id, balance=balance
            )
            self.splitwise_asset = updated_asset
        return self.splitwise_asset

    _deleted_payee = "[DELETED FROM SPLITWISE]"

    def get_new_transactions(
        self,
    ) -> Tuple[List[SplitLunchExpense], List[TransactionObject]]:
        """
        Get Splitwise Transactions that don't exist in Lunch Money

        Also return deleted transaction from LunchMoney

        Returns
        -------
        Tuple[List[SplitLunchExpense], List[TransactionObject]]
            New and Deleted Transactions
        """
        if self.splitwise_asset is None:
            self._raise_splitwise_asset_error()
            raise ValueError("SplitwiseAsset")
        splitlunch_expenses = self.lunchable.get_transactions(
            asset_id=self.splitwise_asset.id,
            start_date=datetime.datetime(1800, 1, 1),
            end_date=datetime.datetime(2300, 12, 31),
        )
        splitlunch_ids = {
            int(item.external_id)
            for item in splitlunch_expenses
            if item.external_id is not None
        }
        splitwise_expenses = self.get_expenses(limit=0)
        splitwise_ids = {item.splitwise_id for item in splitwise_expenses}
        new_ids = splitwise_ids.difference(splitlunch_ids)
        filtered_expenses = self.filter_relevant_splitwise_expenses(
            expenses=splitwise_expenses
        )
        new_expenses = [
            expense for expense in filtered_expenses if expense.splitwise_id in new_ids
        ]
        deleted_transactions = self.get_deleted_transactions(
            splitlunch_expenses=splitlunch_expenses,
            splitwise_transactions=splitwise_expenses,
        )
        return new_expenses, deleted_transactions

    def get_deleted_transactions(
        self,
        splitlunch_expenses: List[TransactionObject],
        splitwise_transactions: List[SplitLunchExpense],
    ) -> List[TransactionObject]:
        """
        Get Splitwise Transactions that exist in Lunch Money but have since been deleted

        Set these transactions to $0.00 and Make a Note

        Parameters
        ----------
        splitlunch_expenses: List[TransactionObject]
        splitwise_transactions: List[SplitLunchExpense]

        Returns
        -------
        List[TransactionObject]
        """
        if self.splitwise_asset is None:
            self._raise_splitwise_asset_error()
            raise ValueError("SplitwiseAsset")
        existing_transactions = {
            int(item.external_id)
            for item in splitlunch_expenses
            if item.external_id is not None
        }
        deleted_ids = {
            item.splitwise_id for item in splitwise_transactions if item.deleted is True
        }
        untethered_transactions = deleted_ids.intersection(existing_transactions)
        transactions_to_delete = [
            tran
            for tran in splitlunch_expenses
            if tran.external_id is not None
            and int(tran.external_id) in untethered_transactions
            and tran.payee != self._deleted_payee
        ]
        return transactions_to_delete

    def refresh_splitwise_transactions(self) -> Dict[str, Any]:
        """
        Import New Splitwise Transactions to Lunch Money

        This function get's all transactions from Splitwise, all transactions from
        your Lunch Money Splitwise account and compares the two.

        Returns
        -------
        List[SplitLunchExpense]
        """
        new_transactions, deleted_transactions = self.get_new_transactions()
        self.splitwise_to_lunchmoney(expenses=new_transactions)
        splitwise_asset = self.update_splitwise_balance()
        self.handle_deleted_transactions(deleted_transactions=deleted_transactions)
        return dict(
            balance=splitwise_asset.balance,
            new=new_transactions,
            deleted=deleted_transactions,
        )

    def handle_deleted_transactions(
        self,
        deleted_transactions: List[TransactionObject],
    ) -> List[Dict[str, Any]]:
        """
        Update Transactions That Exist in Splitwise, but have been deleted in Splitwise

        Parameters
        ----------
        deleted_transactions: List[TransactionObject]

        Returns
        -------
        List[Dict[str, Any]]
        """
        updated_transactions = list()
        for transaction in deleted_transactions:
            update = self.lunchable.update_transaction(
                transaction_id=transaction.id,
                transaction=TransactionUpdateObject(
                    amount=0.00,
                    payee=self._deleted_payee,
                    notes=f"{transaction.payee} || {transaction.amount} || {transaction.notes}",
                ),
            )
            updated_transactions.append(update)
        return updated_transactions

    def _raise_financial_partner_error(self) -> None:
        """
        Raise Errors for Financial Partners
        """
        if self.financial_partner is None:
            raise SplitLunchError(
                "You must designate a financial partner in Splitwise. "
                "This can be done with the partner's Splitwise User ID # "
                "or their email address."
            )

    def _raise_splitwise_asset_error(self) -> None:
        """
        Raise Errors for Splitwise Asset
        """
        raise SplitLunchError(
            "You must create an asset (aka Account) in Lunch Money with "
            "`Splitwise` in the name. There should only be one account "
            "like this."
        )

    def _raise_category_reimbursement_error(self) -> None:
        """
        Raise Errors for Splitwise Asset
        """
        raise SplitLunchError(
            "SplitLunch requires a reimbursement Category. "
            "Make sure you have a category entitled `Reimbursement`. "
            "This category will be excluded from budgeting."
            "Half of split transactions will be created with "
            "this category."
        )



"""
File Path Helper
"""

from pathlib import Path


class FileConfig:
    """
    Configuration Namespace for File Paths
    """

    HOME_DIR = Path.home()
    _file_config_module = Path(__file__).resolve()
    CONFIG_DIR = _file_config_module.parent
    LUNCHMONEY_DIR = CONFIG_DIR.parent
    PROJECT_DIR = LUNCHMONEY_DIR.parent



"""
SplitLunch Data Models
"""

import datetime
from typing import List, Optional

from lunchable.models._base import LunchableModel


class SplitLunchExpense(LunchableModel):
    """
    SplitLunch Object for Splitwise Expenses
    """

    splitwise_id: int
    original_amount: float
    self_paid: bool
    financial_impact: float
    personal_share: float
    description: str
    category: str
    details: Optional[str]
    payment: bool
    date: datetime.datetime
    users: List[int]
    created_at: datetime.datetime
    updated_at: datetime.datetime
    deleted_at: Optional[datetime.datetime]
    deleted: bool



"""
Base Pydantic Object for Containers
"""

from pydantic import BaseModel


class LunchableModel(BaseModel):
    """
    Hashable Pydantic Model
    """

    def __hash__(self):
        """
        Hash Method for Pydantic BaseModels
        """
        return hash((type(self),) + tuple(self.__dict__.values()))



"""
Lunch Money Python Client

This Module Leverages Class Inheritance to distribute API Methods Across a series of
clients. Ultimately, everything inherits from the
lunchable.models.core.LunchMoneyAPIClient class which facilitates interacting with
the API.

For example: to see source code on interactions with the "transactions" API endpoint you
will refer to the TransactionsClient object.
"""

from typing import Optional

from .assets import AssetsClient
from .budgets import BudgetsClient
from .categories import CategoriesClient
from .crypto import CryptoClient
from .plaid_accounts import PlaidAccountsClient
from .recurring_expenses import RecurringExpensesClient
from .tags import TagsClient
from .transactions import TransactionsClient


class LunchMoney(
    AssetsClient,
    BudgetsClient,
    CategoriesClient,
    CryptoClient,
    PlaidAccountsClient,
    RecurringExpensesClient,
    TagsClient,
    TransactionsClient,
):
    """
    Lunch Money Python Client.

    This class facilitates with connections to
    the `Lunch Money Developer API <https://lunchmoney.dev/>`_. Authenticate
    with an Access Token. If an access token isn't provided one will attempt to
    be inherited from a `LUNCHMONEY_ACCESS_TOKEN` environment variable.

    Examples
    --------
    ::

        from lunchable import LunchMoney

        lunch = LunchMoney(access_token="xxxxxxx")
        transactions = lunch.get_transactions()
    """

    def __init__(self, access_token: Optional[str] = None):
        """
        Initialize a Lunch Money object with an Access Token.

        Tries to inherit from the Environment if one isn't provided

        Parameters
        ----------
        access_token: Optional[str]
            Lunchmoney Developer API Access Token
        """
        super(LunchMoney, self).__init__(access_token=access_token)



"""
Lunch Money Python SDK and Associated Objects
"""

from .assets import AssetsObject
from .budgets import BudgetObject
from .categories import CategoriesObject
from .crypto import CryptoObject
from .plaid_accounts import PlaidAccountObject
from .recurring_expenses import RecurringExpensesObject
from .tags import TagsObject
from .transactions import (
    TransactionBaseObject,
    TransactionInsertObject,
    TransactionObject,
    TransactionSplitObject,
    TransactionUpdateObject,
)

__all__ = [
    "AssetsObject",
    "BudgetObject",
    "CategoriesObject",
    "CryptoObject",
    "PlaidAccountObject",
    "RecurringExpensesObject",
    "TagsObject",
    "TransactionBaseObject",
    "TransactionObject",
    "TransactionUpdateObject",
    "TransactionInsertObject",
    "TransactionSplitObject",
]



"""
Lunch Money - Categories

https://lunchmoney.dev/#categories
"""

import datetime
import json
import logging
from typing import List, Optional

from pydantic import Field

from lunchable._config import APIConfig
from lunchable.exceptions import LunchMoneyError
from lunchable.models._base import LunchableModel
from lunchable.models._core import LunchMoneyAPIClient

logger = logging.getLogger(__name__)


class ModelCreateCategory(LunchableModel):
    """
    https://lunchmoney.dev/#create-category
    """

    name: str
    description: Optional[str]
    is_income: Optional[bool] = False
    exclude_from_budget: Optional[bool] = False
    exclude_from_totals: Optional[bool] = False


class CategoryChild(LunchableModel):
    """
    Child Entry on the Category Object
    """

    id: int
    name: str = Field(min_length=1, max_length=40)
    description: Optional[str] = Field(max_length=140)
    created_at: Optional[datetime.datetime]


class CategoriesObject(LunchableModel):
    """
    Lunch Money Spending Categories

    https://lunchmoney.dev/#categories-object
    """

    _name_description = "The name of the category. Must be between 1 and 40 characters."
    _description_description = (
        "The description of the category. Must not exceed 140 characters."
    )
    _is_income_description = (
        "If true, the transactions in this category will be treated as income."
    )
    _exclude_from_budget_description = """
    If true, the transactions in this category will be excluded from the budget.
    """
    _exclude_from_totals_description = """
    If true, the transactions in this category will be excluded from totals.
    """
    _updated_at_description = """
    The date and time of when the category was last updated (in the ISO 8601 extended format).
    """
    _created_at_description = """
    The date and time of when the category was created (in the ISO 8601 extended format).
    """
    _is_group_description = """
    If true, the category is a group that can be a parent to other categories.
    """
    _group_id_description = """
    The ID of a category group (or null if the category doesn't belong to a category group).
    """
    _children_description = (
        "For category groups, this will populate with the "
        "categories nested within and include id, name, "
        "description and created_at fields."
    )

    id: int = Field(description="A unique identifier for the category.")
    name: str = Field(min_length=1, max_length=40, description=_name_description)
    description: Optional[str] = Field(
        max_length=140, description=_description_description
    )
    is_income: str = Field(description=_is_income_description)
    exclude_from_budget: bool = Field(description=_exclude_from_budget_description)
    exclude_from_totals: bool = Field(description=_exclude_from_totals_description)
    updated_at: Optional[datetime.datetime] = Field(description=_updated_at_description)
    created_at: Optional[datetime.datetime] = Field(description=_created_at_description)
    is_group: bool = Field(description=_is_group_description)
    group_id: Optional[int] = Field(description=_group_id_description)
    children: Optional[List[CategoryChild]] = Field(description=_children_description)


class _CategoriesParamsPut(LunchableModel):
    """
    https://lunchmoney.dev/#update-category
    """

    name: Optional[str] = Field(min_length=1, max_length=40)
    description: Optional[str] = Field(max_length=140)
    is_income: Optional[bool]
    exclude_from_budget: Optional[bool]
    exclude_from_totals: Optional[bool]
    group_id: Optional[int]


class _CategoriesParamsPost(LunchableModel):
    """
    https://lunchmoney.dev/#create-category-group
    """

    name: str
    description: Optional[str]
    is_income: Optional[bool] = False
    exclude_from_budget: Optional[bool] = False
    exclude_from_totals: Optional[bool] = False
    category_ids: Optional[List[int]]
    new_categories: Optional[List[str]]


class _CategoriesAddParamsPost(LunchableModel):
    """
    https://lunchmoney.dev/#add-to-category-group
    """

    category_ids: Optional[List[int]]
    new_categories: Optional[List[str]]


class CategoriesClient(LunchMoneyAPIClient):
    """
    Lunch Money Categories Interactions
    """

    def get_categories(self) -> List[CategoriesObject]:
        """
        Get Spending categories

        Use this endpoint to get a list of all categories associated with the user's account.
        https://lunchmoney.dev/#get-all-categories

        Returns
        -------
        List[CategoriesObject]
        """
        response_data = self._make_request(
            method=self.Methods.GET, url_path=APIConfig.LUNCHMONEY_CATEGORIES
        )
        categories = response_data["categories"]
        category_objects = [CategoriesObject(**item) for item in categories]
        return category_objects

    def insert_category(
        self,
        name: str,
        description: Optional[str] = None,
        is_income: Optional[bool] = False,
        exclude_from_budget: Optional[bool] = False,
        exclude_from_totals: Optional[bool] = False,
    ) -> int:
        """
        Create a Spending Category

        Use this to create a single category
        https://lunchmoney.dev/#create-category

        Parameters
        ----------
        name: str
            Name of category. Must be between 1 and 40 characters.
        description: Optional[str]
            Description of category. Must be less than 140 categories. Defaults to None.
        is_income: Optional[bool]
            Whether or not transactions in this category should be treated as income.
            Defaults to False.
        exclude_from_budget: Optional[bool]
            Whether or not transactions in this category should be excluded from budgets.
            Defaults to False.
        exclude_from_totals: Optional[bool]
            Whether or not transactions in this category should be excluded from
            calculated totals. Defaults to False.

        Returns
        -------
        int
            ID of the newly created category
        """
        category_body = ModelCreateCategory(
            name=name,
            description=description,
            is_income=is_income,
            exclude_from_budget=exclude_from_budget,
            exclude_from_totals=exclude_from_totals,
        ).dict(exclude_none=True)
        response_data = self._make_request(
            method=self.Methods.POST,
            url_path=APIConfig.LUNCHMONEY_CATEGORIES,
            payload=category_body,
        )
        return response_data["category_id"]

    def get_category(self, category_id: int) -> CategoriesObject:
        """
        Get single category

        Use this endpoint to get hydrated details on a single category. Note that if
        this category is part of a category group, its properties (is_income,
        exclude_from_budget, exclude_from_totals) will inherit from the category group.

        https://lunchmoney.dev/#get-single-category

        Parameters
        ----------
        category_id : int
            Id of the Lunch Money Category

        Returns
        -------
        CategoriesObject
        """
        response_data = self._make_request(
            method=self.Methods.GET,
            url_path=[APIConfig.LUNCHMONEY_CATEGORIES, category_id],
        )
        return CategoriesObject(**response_data)

    def remove_category(self, category_id: int) -> bool:
        """
        Delete a single category

        Use this endpoint to delete a single category or category group. This will
        only work if there are no dependencies, such as existing budgets for the
        category, categorized transactions, categorized recurring items, etc. If
        there are dependents, this endpoint will return what the dependents are
        and how many there are.

        https://lunchmoney.dev/#delete-category

        Parameters
        ----------
        category_id : int
            Id of the Lunch Money Category

        Returns
        -------
        bool
        """
        response_data = self._make_request(
            method=self.Methods.DELETE,
            url_path=[APIConfig.LUNCHMONEY_CATEGORIES, category_id],
        )
        if response_data is not True:
            raise LunchMoneyError(
                f"That Category ({category_id}) has Dependents: "
                f"{json.dumps(response_data, indent=4)}"
            )
        return response_data

    def remove_category_force(self, category_id: int) -> bool:
        """
        Forcefully delete a single category

        Use this endpoint to force delete a single category or category group and
        along with it, disassociate the category from any transactions, recurring
        items, budgets, etc.

        Note: it is best practice to first try the Delete Category endpoint to ensure
        you don't accidentally delete any data. Disassociation/deletion of the data
        arising from this endpoint is irreversible!

        https://lunchmoney.dev/#force-delete-category

        Parameters
        ----------
        category_id : int
            Id of the Lunch Money Category

        Returns
        -------
        bool
        """
        response_data = self._make_request(
            method=self.Methods.DELETE,
            url_path=[APIConfig.LUNCHMONEY_CATEGORIES, category_id, "force"],
        )
        return response_data

    def update_category(
        self,
        category_id: int,
        name: Optional[str] = None,
        description: Optional[str] = None,
        is_income: Optional[bool] = None,
        exclude_from_budget: Optional[bool] = None,
        exclude_from_totals: Optional[bool] = None,
        group_id: Optional[int] = None,
    ) -> bool:
        """
        Update a single category

        Use this endpoint to update the properties for a single category or category group

        https://lunchmoney.dev/#update-category

        Parameters
        ----------
        category_id : int
            Id of the Lunch Money Category
        name: str
            Name of category. Must be between 1 and 40 characters.
        description: Optional[str]
            Description of category. Must be less than 140 categories. Defaults to None.
        is_income: Optional[bool]
            Whether or not transactions in this category should be treated as income.
            Defaults to False.
        exclude_from_budget: Optional[bool]
            Whether or not transactions in this category should be excluded from budgets.
            Defaults to False.
        exclude_from_totals: Optional[bool]
            Whether or not transactions in this category should be excluded from
            calculated totals. Defaults to False.
        group_id: Optional[int]
            For a category, set the group_id to include it in a category group

        Returns
        -------
        bool
        """
        payload = _CategoriesParamsPut(
            name=name,
            description=description,
            is_income=is_income,
            exclude_from_budget=exclude_from_budget,
            exclude_from_totals=exclude_from_totals,
            group_id=group_id,
        ).dict(exclude_none=True)
        response_data = self._make_request(
            method=self.Methods.PUT,
            url_path=[APIConfig.LUNCHMONEY_CATEGORIES, category_id],
            payload=payload,
        )
        return response_data

    def insert_category_group(
        self,
        name: str,
        description: Optional[str] = None,
        is_income: Optional[bool] = False,
        exclude_from_budget: Optional[bool] = False,
        exclude_from_totals: Optional[bool] = False,
        category_ids: Optional[List[int]] = None,
        new_categories: Optional[List[str]] = None,
    ) -> int:
        """
        Create a Spending Category Group

        Use this endpoint to create a single category group
        https://lunchmoney.dev/#create-category-group

        Parameters
        ----------
        name: str
            Name of category. Must be between 1 and 40 characters.
        description: Optional[str]
            Description of category. Must be less than 140 categories. Defaults to None.
        is_income: Optional[bool]
            Whether or not transactions in this category should be treated as income.
            Defaults to False.
        exclude_from_budget: Optional[bool]
            Whether or not transactions in this category should be excluded from budgets.
            Defaults to False.
        exclude_from_totals: Optional[bool]
            Whether or not transactions in this category should be excluded from
            calculated totals. Defaults to False.
        category_ids: Optional[List[int]]
            Array of category_id to include in the category group.
        new_categories: Optional[List[str]]
            Array of strings representing new categories to create and subsequently
            include in the category group.

        Returns
        -------
        int
            ID of the newly created category group
        """
        payload = _CategoriesParamsPost(
            name=name,
            description=description,
            is_income=is_income,
            exclude_from_budget=exclude_from_budget,
            exclude_from_totals=exclude_from_totals,
            category_ids=category_ids,
            new_categories=new_categories,
        ).dict(exclude_none=True)
        response_data = self._make_request(
            method=self.Methods.POST,
            url_path=[APIConfig.LUNCHMONEY_CATEGORIES, "group"],
            payload=payload,
        )
        return response_data["category_id"]

    def insert_into_category_group(
        self,
        category_group_id: int,
        category_ids: Optional[List[int]] = None,
        new_categories: Optional[List[str]] = None,
    ) -> CategoriesObject:
        """
        Add to a Category Group

        Use this endpoint to add categories (either existing or new) to a single
        category group

        https://lunchmoney.dev/#add-to-category-group

        Parameters
        ----------
        category_group_id: int
            Id of the Lunch Money Category Group
        category_ids: Optional[List[int]]
            Array of category_id to include in the category group.
        new_categories: Optional[List[str]]
            Array of strings representing new categories to create and subsequently
            include in the category group.

        Returns
        -------
        CategoriesObject
        """
        payload = _CategoriesAddParamsPost(
            category_ids=category_ids, new_categories=new_categories
        ).dict(exclude_none=True)
        response_data = self._make_request(
            method=self.Methods.POST,
            url_path=[
                APIConfig.LUNCHMONEY_CATEGORIES,
                "group",
                category_group_id,
                "add",
            ],
            payload=payload,
        )
        return CategoriesObject(**response_data)



"""
Pushover Notifications via lunchable
"""

import datetime
import logging
from base64 import b64decode
from json import loads
from os import getenv
from textwrap import dedent
from time import sleep
from typing import Any, Dict, List, Optional

import requests

from lunchable import LunchMoney
from lunchable.models import AssetsObject, PlaidAccountObject, TransactionObject

logger = logging.getLogger(__name__)


class PushLunchError(Exception):
    """
    PushLunch Exception
    """

    pass


class PushLunch:
    """
    Lunch Money Pushover Notifications via Lunchable
    """

    pushover_endpoint = "https://api.pushover.net/1/messages.json"

    def __init__(
        self,
        user_key: Optional[str] = None,
        app_token: Optional[str] = None,
        lunchmoney_access_token: Optional[str] = None,
        lunchable_client: Optional[LunchMoney] = None,
    ):
        """
        Initialize

        Parameters
        ----------
        user_key : Optional[str]
            Pushover User Key. Will attempt to inherit from `PUSHOVER_USER_KEY`  environment
            variable if none defined
        app_token: Optional[str]
            Pushover app token, will attempt to inherit from `PUSHOVER_APP_TOKEN`  environment
            variable. If no token available, the official lunchable app token will be provided
        lunchmoney_access_token: Optional[str]
            LunchMoney Access Token. Will be inherited from `LUNCHMONEY_ACCESS_TOKEN`
            environment variable.
        lunchable_client: Optional[LunchMoney]
            lunchable client to use. One will be created if none provided.
        """
        self.session = requests.Session()
        self.session.headers.update({"Content-Type": "application/json"})

        _courtesy_token = b"YXpwMzZ6MjExcWV5OGFvOXNicWF0cmdraXc4aGVz"
        if app_token is None:
            app_token = getenv("PUSHOVER_APP_TOKEN", None)
        token = app_token or b64decode(_courtesy_token).decode("utf-8")
        user_key = user_key or getenv("PUSHOVER_USER_KEY", None)
        if user_key in [None, ""]:
            raise PushLunchError(
                "You must provide a Pushover User Key or define it with "
                "a `PUSHOVER_USER_KEY` environment variable"
            )
        self._params = dict(user=user_key, token=token)
        self.lunchable = lunchable_client or LunchMoney(
            access_token=lunchmoney_access_token
        )
        self.asset_mapping = self._get_assets()
        self.category_mapping = self._get_categories()
        self.notified_transactions: List[int] = list()

    def send_notification(
        self,
        message: str,
        attachment: Optional[object] = None,
        device: Optional[str] = None,
        title: Optional[str] = None,
        url: Optional[str] = None,
        url_title: Optional[str] = None,
        priority: Optional[int] = None,
        sound: Optional[str] = None,
        timestamp: Optional[str] = None,
        html: bool = False,
    ) -> requests.Response:
        """
        Send a Pushover Notification

        Parameters
        ----------
        message: Optional[str]
            your message
        attachment: Optional[object]
            an image attachment to send with the message; see attachments for more information
            on how to upload files
        device: Optional[str]
            your user's device name to send the message directly to that device, rather than
            all of the user's devices (multiple devices may be separated by a comma)
        title: Optional[str]
            your message's title, otherwise your app's name is used
        url: Optional[str]
            a supplementary URL to show with your message
        url_title: Optional[str]
            a title for your supplementary URL, otherwise just the URL is shown
        priority: Optional[int]
            send as -2 to generate no notification/alert, -1 to always send as a quiet
            notification, 1 to display as high-priority and bypass the user's quiet hours,
            or 2 to also require confirmation from the user
        sound: Optional[str]
            the name of one of the sounds supported by device clients to override the
            user's default sound choice
        timestamp: Optional[str]
            a Unix timestamp of your message's date and time to display to the user, rather
            than the time your message is received by our API
        html: Union[None, 1]
            Pass 1 if message contains HTML contents

        Returns
        -------
        requests.Response
        """
        html_param = 1 if html not in [None, False] else None
        params_dict = dict(
            message=message,
            attachment=attachment,
            device=device,
            title=title,
            url=url,
            url_title=url_title,
            priority=priority,
            sound=sound,
            timestamp=timestamp,
            html=html_param,
        )
        params: Dict[str, Any] = {
            key: value for key, value in params_dict.items() if value is not None
        }
        params.update(self._params)
        response = self.session.post(url=self.pushover_endpoint, params=params)
        response.raise_for_status()
        return response

    def post_transaction(
        self, transaction: TransactionObject
    ) -> Optional[Dict[str, Any]]:
        """
        Post a Lunch Money Transaction as a Pushover Notification

        Assuming the instance of the
        class hasn't already posted this particular notification

        Parameters
        ----------
        transaction: TransactionObject

        Returns
        -------
        Dict[str, Any]
        """
        if transaction.id in self.notified_transactions:
            return None
        if transaction.category_id is None:
            category = "N/A"
        else:
            category = self.category_mapping[transaction.category_id]
        account_id = transaction.plaid_account_id or transaction.asset_id
        assert account_id is not None
        transaction_formatted = dedent(
            f"""
        <b>Payee:</b> <i>{transaction.payee}</i>
        <b>Amount:</b> <i>{self._format_float(transaction.amount)}</i>
        <b>Date:</b> <i>{transaction.date.strftime("%A %B %-d, %Y")}</i>
        <b>Category:</b> <i>{category}</i>
        <b>Account:</b> <i>{self.asset_mapping[account_id]}</i>
        """
        ).strip()
        if transaction.currency is not None:
            transaction_formatted += (
                f"\n<b>Currency:</b> <i>{transaction.currency.upper()}</i>"
            )
        if transaction.status is not None:
            transaction_formatted += (
                f"\n<b>Status:</b> <i>{transaction.status.title()}</i>"
            )
        if transaction.notes is not None:
            note = f"<b>Notes:</b> <i>{transaction.notes}</i>"
            transaction_formatted += f"\n{note}"
        if transaction.status == "uncleared":
            url = (
                '<a href="https://my.lunchmoney.app/transactions/'
                f'{transaction.date.year}/{transaction.date.strftime("%m")}?status=unreviewed">'
                "<b>Uncleared Transactions from this Period</b></a>"
            )
            transaction_formatted += f"\n\n{url}"

        response = self.send_notification(
            message=transaction_formatted, title="Lunch Money Transaction", html=True
        )
        self.notified_transactions.append(transaction.id)
        return loads(response.content)

    def _get_assets(self) -> Dict[int, str]:
        """
        Get Mapping Of Asset ID -> Asset Name

        Returns
        -------
        Dict[int, str]
        """
        manual_assets = self.lunchable.get_assets()
        plaid_account = self.lunchable.get_plaid_accounts()
        assets = [*manual_assets, *plaid_account]
        asset_mapping = dict()
        for account in assets:
            if isinstance(account, AssetsObject):
                if account.display_name is None:
                    name = account.name
                else:
                    name = account.display_name
                asset_mapping[account.id] = name
            elif isinstance(account, PlaidAccountObject):
                asset_mapping[account.id] = account.name
        return asset_mapping

    def _get_categories(self) -> Dict[int, str]:
        """
        Get Mapping Of Category ID -> Category Name

        Returns
        -------
        Dict[int, str]
        """
        categories = self.lunchable.get_categories()
        category_mapping = dict()
        for category in categories:
            category_mapping[category.id] = category.name
        return category_mapping

    @classmethod
    def _format_float(cls, amount: float) -> str:
        """
        Format Floats to be pleasant and human readable

        Parameters
        ----------
        amount: float
            Float Amount to be converted into a string

        Returns
        -------
        str
        """
        if amount < 0:
            float_string = "$ ({:,.2f})".format(float(amount)).replace("-", "")
        else:
            float_string = "$ {:,.2f}".format(float(amount))
        return float_string

    def _get_uncleared_transactions(
        self,
        start_date: Optional[datetime.datetime] = None,
        end_date: Optional[datetime.datetime] = None,
    ) -> List[TransactionObject]:
        """
        Get Uncleared Transactions

        Returns
        -------
        List[TransactionObject]
        """
        uncleared_transactions = self.lunchable.get_transactions(
            start_date=start_date, end_date=end_date, status="uncleared"
        )
        return uncleared_transactions

    def notify_uncleared_transactions(
        self, continuous: bool = False, interval: Optional[int] = None
    ) -> List[TransactionObject]:
        """
        Get the Current Period's Uncleared Transactions and Send a Notification for each

        Parameters
        ----------
        continuous : bool
            Whether to continuously check for more uncleared transactions,
            waiting a fixed amount in between checks.
        interval: Optional[int]
            Sleep Interval in Between Tries - only applies if `continuous` is set.
            Defaults to 60 (minutes). Cannot be less than 5 (minutes)

        Returns
        -------
        List[TransactionObject]
        """
        if interval is None:
            interval = 60
        if continuous is True and interval < 5:
            logger.warning(
                "Check interval cannot be less than 5 minutes. Defaulting to 5."
            )
            interval = 5
        if continuous is True:
            logger.info("Continuous Notifications Enabled. Beginning PushLunch.")

        uncleared_transactions = list()
        continuous_search = True

        while continuous_search is True:
            found_transactions = len(self.notified_transactions)
            uncleared_transactions += self._get_uncleared_transactions()
            for transaction in uncleared_transactions:
                self.post_transaction(transaction=transaction)
            if continuous is True:
                notified = len(self.notified_transactions)
                new_transactions = notified - found_transactions
                logger.info(
                    "%s new transactions pushed. %s total.", new_transactions, notified
                )
                sleep(interval * 60)
            else:
                continuous_search = False

        return uncleared_transactions



"""
SplitLunch Configuration Helpers
"""


class SplitLunchConfig:
    """
    Configuration Namespace for SplitLunch
    """

    splitlunch_tag: str = "SplitLunch"
    splitwise_tag: str = "Splitwise"
    splitlunch_import_tag: str = "SplitLunchImport"
    splitlunch_direct_import_tag: str = "SplitLunchDirectImport"



"""
Lunch Money - Crypto

https://lunchmoney.dev/#crypto
"""

import datetime
import logging
from typing import List, Optional

from pydantic import Field

from lunchable._config import APIConfig
from lunchable.models._base import LunchableModel
from lunchable.models._core import LunchMoneyAPIClient

logger = logging.getLogger(__name__)


class CryptoObject(LunchableModel):
    """
    Crypto Asset Object

    https://lunchmoney.dev/#crypto-object
    """

    _id_description = (
        "Unique identifier for a manual crypto account (no ID for synced accounts)"
    )
    _zabo_account_id_description = """
    Unique identifier for a synced crypto account (no ID for manual accounts,
    multiple currencies may have the same zabo_account_id)
    """
    _source_description = """
    `synced` (this account is synced via a wallet, exchange, etc.) or `manual` (this account
    balance is managed manually)
    """
    _display_name_description = "Display name of the crypto asset (as set by user)"
    _balance_as_of_description = """
    Date/time the balance was last updated in ISO 8601 extended format
    """
    _status_description = (
        "The current status of the crypto account. Either active or in error."
    )
    _created_at_description = (
        "Date/time the asset was created in ISO 8601 extended format"
    )

    id: int = Field(description=_id_description)
    zabo_account_id: Optional[int] = Field(description=_zabo_account_id_description)
    source: str = Field(description=_source_description)
    name: str = Field(description="Name of the crypto asset")
    display_name: Optional[str] = Field(description=_display_name_description)
    balance: float = Field(description="Current balance")
    balance_as_of: Optional[datetime.datetime] = Field(
        description=_balance_as_of_description
    )
    currency: Optional[str] = Field(description="Abbreviation for the cryptocurrency")
    status: Optional[str] = Field(description=_status_description)
    institution_name: str = Field(description="Name of provider holding the asset")
    created_at: datetime.datetime = Field(description=_created_at_description)


class CryptoParamsPut(LunchableModel):
    """
    https://lunchmoney.dev/#update-manual-crypto-asset
    """

    name: Optional[str]
    display_name: Optional[str]
    institution_name: Optional[str]
    balance: Optional[float]
    currency: Optional[str]


class CryptoClient(LunchMoneyAPIClient):
    """
    Lunch Money Tag Interactions
    """

    def get_crypto(self) -> List[CryptoObject]:
        """
        Get Crypto Assets

        Use this endpoint to get a list of all cryptocurrency assets associated
        with the user's account. Both crypto balances from synced and manual
        accounts will be returned.

        https://lunchmoney.dev/#get-all-crypto

        Returns
        -------
        List[CryptoObject]
        """
        response_data = self._make_request(
            method=self.Methods.GET, url_path=APIConfig.LUNCHMONEY_CRYPTO
        )
        crypto_data = response_data["crypto"]
        crypto_objects = [CryptoObject(**item) for item in crypto_data]
        return crypto_objects

    def update_crypto(
        self,
        crypto_id: int,
        name: Optional[str] = None,
        display_name: Optional[str] = None,
        institution_name: Optional[str] = None,
        balance: Optional[float] = None,
        currency: Optional[str] = None,
    ) -> CryptoObject:
        """
        Update a Manual Crypto Asset

        Use this endpoint to update a single manually-managed crypto asset (does not include
        assets received from syncing with your wallet/exchange/etc). These are denoted by
        source: manual from the GET call above.

        https://lunchmoney.dev/#update-manual-crypto-asset

        Parameters
        ----------
        crypto_id: int
            ID of the crypto asset to update
        name: Optional[str]
            Official or full name of the account. Max 45 characters
        display_name: Optional[str]
            Display name for the account. Max 25 characters
        institution_name: Optional[str]
            Name of provider that holds the account. Max 50 characters
        balance: Optional[float]
            Numeric value of the current balance of the account. Do not include any
            special characters aside from a decimal point!
        currency: Optional[str]
            Cryptocurrency that is supported for manual tracking in our database

        Returns
        -------
        CryptoObject
        """
        crypto_body = CryptoParamsPut(
            name=name,
            display_name=display_name,
            institution_name=institution_name,
            balance=balance,
            currency=currency,
        ).dict(exclude_none=True)
        response_data = self._make_request(
            method=self.Methods.PUT,
            url_path=[
                APIConfig.LUNCHMONEY_CRYPTO,
                APIConfig.LUNCHMONEY_CRYPTO_MANUAL,
                crypto_id,
            ],
            payload=crypto_body,
        )
        crypto = CryptoObject(**response_data)
        return crypto



"""
Lunchmoney Exceptions
"""

from requests.exceptions import HTTPError


class LunchMoneyError(Exception):
    """
    Base Exception for Lunch Money
    """


class EnvironmentVariableError(LunchMoneyError, EnvironmentError):
    """
    Lunch Money Missing Environment Variable Error
    """


class LunchMoneyHTTPError(LunchMoneyError, HTTPError):
    """
    Lunch Money HTTP Error
    """


class LunchMoneyImportError(LunchMoneyError, ImportError):
    """
    Lunch Money Import Error
    """



"""
Lunch Money - Plaid Accounts

https://lunchmoney.dev/#plaid-accounts
"""

import datetime
import logging
from typing import List, Optional

from pydantic import Field

from lunchable._config import APIConfig
from lunchable.models._base import LunchableModel
from lunchable.models._core import LunchMoneyAPIClient

logger = logging.getLogger(__name__)


class PlaidAccountObject(LunchableModel):
    """
    Assets synced from Plaid

    Similar to AssetObjects, these accounts are linked to remote sources in Plaid.

    https://lunchmoney.dev/#plaid-accounts-object
    """

    _date_linked_description = (
        "Date account was first linked in ISO 8601 extended format"
    )
    _name_description = """
    Name of the account. Can be overridden by the user. Field is originally set by Plaid")
    """
    _type_description = """
    Primary type of account. Typically one of: [credit, depository, brokerage, cash,
    loan, investment]. This field is set by Plaid and cannot be altered.
    """
    _subtype_description = """
    Optional subtype name of account. This field is set by Plaid and cannot be altered
    """
    _mask_description = """
    Mask (last 3 to 4 digits of account) of account. This field is set by
    Plaid and cannot be altered
    """
    _institution_name_description = """
    Name of institution associated with account. This field is set by
    Plaid and cannot be altered
    """
    _status_description = """
    Denotes the current status of the account within Lunch Money. Must be one of:
    active (Account is active and in good state),
    inactive (Account marked inactive from user. No transactions fetched or
    balance update for this account),
    relink (Account needs to be relinked with Plaid),
    syncing (Account is awaiting first import of transactions),
    error (Account is in error with Plaid),
    not found (Account is in error with Plaid),
    not supported (Account is in error with Plaid)
    """
    _last_import_description = """
    Date of last imported transaction in ISO 8601 extended format (not necessarily
    date of last attempted import)
    """
    _balance_description = """
    Current balance of the account in numeric format to 4 decimal places. This field is
    set by Plaid and cannot be altered
    """
    _currency_description = """
    Currency of account balance in ISO 4217 format. This field is set by Plaid
    and cannot be altered
    """
    _balance_last_update_description = """
    Date balance was last updated in ISO 8601 extended format. This field is set
    by Plaid and cannot be altered
    """
    _limit_description = """
    Optional credit limit of the account. This field is set by Plaid and cannot be altered
    """

    id: int = Field(description="Unique identifier of Plaid account")
    date_linked: datetime.date = Field(description=_date_linked_description)
    name: str = Field(description=_name_description)
    type: str = Field(description=_type_description)
    subtype: str = Field(description=_subtype_description)
    mask: str = Field(description=_mask_description)
    institution_name: str = Field(description=_institution_name_description)
    status: str = Field(description=_status_description)
    last_import: Optional[datetime.datetime] = Field(
        description=_last_import_description
    )
    balance: Optional[float] = Field(description=_balance_description)
    currency: str = Field(description=_currency_description)
    balance_last_update: datetime.datetime = Field(
        description=_balance_last_update_description
    )
    limit: Optional[int] = Field(description=_limit_description)


class PlaidAccountsClient(LunchMoneyAPIClient):
    """
    Lunch Money Plaid Accounts Interactions
    """

    def get_plaid_accounts(self) -> List[PlaidAccountObject]:
        """
        Get Plaid Synced Assets

        Get a list of all synced Plaid accounts associated with the user's account.

        Plaid Accounts are individual bank accounts that you have linked to Lunch Money via Plaid.
        You may link one bank but one bank might contain 4 accounts. Each of these
        accounts is a Plaid Account. (https://lunchmoney.dev/#plaid-accounts-object)

        Returns
        -------
        List[PlaidAccountObject]
        """
        response_data = self._make_request(
            method=self.Methods.GET, url_path=APIConfig.LUNCHMONEY_PLAID_ACCOUNTS
        )
        accounts = response_data.get(APIConfig.LUNCHMONEY_PLAID_ACCOUNTS)
        account_objects = [PlaidAccountObject(**item) for item in accounts]
        return account_objects



"""
Lunch Money - Transactions

https://lunchmoney.dev/#transactions
"""

import datetime
import logging
from enum import Enum
from typing import Any, Dict, List, Optional, Union

from pydantic import Field

from lunchable import LunchMoneyError
from lunchable._config import APIConfig
from lunchable.models._base import LunchableModel
from lunchable.models._core import LunchMoneyAPIClient
from lunchable.models.tags import TagsObject

logger = logging.getLogger(__name__)


class TransactionBaseObject(LunchableModel):
    """
    Base Model For All Transactions to Inherit From
    """

    pass


class TransactionInsertObject(TransactionBaseObject):
    """
    Object For Creating New Transactions

    https://lunchmoney.dev/#insert-transactions
    """

    _date_description = """
    Must be in ISO 8601 format (YYYY-MM-DD).
    """
    _amount_description = """
    Numeric value of amount. i.e. $4.25 should be denoted as 4.25.
    """
    _category_id_description = """
    Unique identifier for associated category_id. Category must be associated with
    the same account and must not be a category group.
    """
    _currency_description = """
    Three-letter lowercase currency code in ISO 4217 format. The code sent must exist
    in our database. Defaults to user account's primary currency.
    """
    _asset_id_description = """
    Unique identifier for associated asset (manually-managed account). Asset must be
    associated with the same account.
    """
    _recurring_id = """
    Unique identifier for associated recurring expense. Recurring expense must be associated
    with the same account.
    """
    _status_description = """
    Must be either cleared or uncleared. If recurring_id is provided, the status will
    automatically be set to recurring or recurring_suggested depending on the type of
    recurring_id. Defaults to uncleared.
    """
    _external_id_description = """
    User-defined external ID for transaction. Max 75 characters. External IDs must be
    unique within the same asset_id.
    """
    _tags_description = """
    Passing in a number will attempt to match by ID. If no matching tag ID is found, an error
    will be thrown. Passing in a string will attempt to match by string. If no matching tag
    name is found, a new tag will be created.
    """

    class StatusEnum(str, Enum):
        """
        Status Options, must be "cleared" or "uncleared"
        """

        cleared = "cleared"
        uncleared = "uncleared"

    date: datetime.date = Field(description=_date_description)
    amount: float = Field(description=_amount_description)
    category_id: Optional[int] = Field(description=_category_id_description)
    payee: Optional[str] = Field(description="Max 140 characters", max_length=140)
    currency: Optional[str] = Field(description=_currency_description, max_length=3)
    asset_id: Optional[int] = Field(description=_asset_id_description)
    recurring_id: Optional[int] = Field(description=_recurring_id)
    notes: Optional[str] = Field(description="Max 350 characters", max_length=350)
    status: Optional[StatusEnum] = Field(description=_status_description)
    external_id: Optional[str] = Field(
        description=_external_id_description, max_length=75
    )
    tags: Optional[List[Union[str, int]]] = Field(description=_tags_description)


class TransactionUpdateObject(TransactionBaseObject):
    """
    Object For Updating Existing Transactions

    https://lunchmoney.dev/#update-transaction
    """

    _date_description = """
    Must be in ISO 8601 format (YYYY-MM-DD).
    """
    _category_id_description = """
    Unique identifier for associated category_id. Category must be associated
    with the same account and must not be a category group.
    """
    _amount_description = """
    You may only update this if this transaction was not created from an automatic
    import, i.e. if this transaction is not associated with a plaid_account_id
    """
    _currency_description = """
    You may only update this if this transaction was not created from an automatic
    import, i.e. if this transaction is not associated with a plaid_account_id.
    Defaults to user account's primary currency.
    """
    _asset_id_description = """
    Unique identifier for associated asset (manually-managed account). Asset must be
    associated with the same account. You may only update this if this transaction was
    not created from an automatic import, i.e. if this transaction is not associated
    with a plaid_account_id
    """
    _recurring_id_description = """
    Unique identifier for associated recurring expense. Recurring expense must
    be associated with the same account.
    """
    _status_description = """
    Must be either cleared or uncleared. Defaults to uncleared If recurring_id is
    provided, the status will automatically be set to recurring or recurring_suggested
    depending on the type of recurring_id. Defaults to uncleared.
    """
    _external_id_description = """
    User-defined external ID for transaction. Max 75 characters. External IDs must be
    unique within the same asset_id. You may only update this if this transaction was
    not created from an automatic import, i.e. if this transaction is not associated
    with a plaid_account_id
    """
    _tags_description = """
    Passing in a number will attempt to match by ID. If no matching tag ID is found,
    an error will be thrown. Passing in a string will attempt to match by string.
    If no matching tag name is found, a new tag will be created.
    """

    class StatusEnum(str, Enum):
        """
        Status Options, must be "cleared" or "uncleared"
        """

        cleared = "cleared"
        uncleared = "uncleared"

    date: Optional[datetime.date] = Field(description=_date_description)
    category_id: Optional[int] = Field(description=_category_id_description)
    payee: Optional[str] = Field(description="Max 140 characters", max_length=140)
    amount: Optional[float] = Field(description=_amount_description)
    currency: Optional[str] = Field(description=_currency_description)
    asset_id: Optional[int] = Field(description=_asset_id_description)
    recurring_id: Optional[int] = Field(description=_recurring_id_description)
    notes: Optional[str] = Field(description="Max 350 characters", max_length=350)
    status: Optional[StatusEnum] = Field(description=_status_description)
    external_id: Optional[str] = Field(description=_external_id_description)
    tags: Optional[List[Union[int, str]]] = Field(description=_tags_description)


class TransactionSplitObject(TransactionBaseObject):
    """
    Object for Splitting Transactions

    https://lunchmoney.dev/#split-object
    """

    _date_description = "Must be in ISO 8601 format (YYYY-MM-DD)."
    _category_id_description = """
    Unique identifier for associated category_id. Category must be associated
    with the same account.
    """
    _notes_description = "Transaction Split Notes."
    _amount_description = """
    Individual amount of split. Currency will inherit from parent transaction. All
    amounts must sum up to parent transaction amount.
    """

    date: datetime.date = Field(description=_date_description)
    category_id: int = Field(description=_category_id_description)
    notes: Optional[str] = Field(description=_notes_description)
    amount: float = Field(description=_amount_description)


class TransactionObject(TransactionBaseObject):
    """
    Universal Lunch Money Transaction Object

    https://lunchmoney.dev/#transaction-object
    """

    _amount_description = """
    Amount of the transaction in numeric format to 4 decimal places
    """
    _payee_description = """
    Name of payee If recurring_id is not null, this field will show the payee
    of associated recurring expense instead of the original transaction payee
    """
    _currency_description = """
    Three-letter lowercase currency code of the transaction in ISO 4217 format
    """
    _notes_description = """
    User-entered transaction notes If recurring_id is not null, this field will
    be description of associated recurring expense
    """
    _category_description = """
    Unique identifier of associated category (see Categories)
    """
    _asset_id_description = """
    Unique identifier of associated manually-managed account (see Assets)
    Note: plaid_account_id and asset_id cannot both exist for a transaction
    """
    _plaid_account_id_description = """
    Unique identifier of associated Plaid account (see Plaid Accounts) Note:
    plaid_account_id and asset_id cannot both exist for a transaction
    """
    _status_description = """
    One of the following: cleared: User has reviewed the transaction uncleared:
    User has not yet reviewed the transaction recurring: Transaction is linked
    to a recurring expense recurring_suggested: Transaction is listed as a
    suggested transaction for an existing recurring expense. User intervention
    is required to change this to recurring.
    """
    _parent_id_description = """
    Exists if this is a split transaction. Denotes the transaction ID of the original
    transaction. Note that the parent transaction is not returned in this call.
    """
    _is_group_description = """
    True if this transaction represents a group of transactions. If so, amount
    and currency represent the totalled amount of transactions bearing this
    transaction’s id as their group_id. Amount is calculated based on the
    user’s primary currency.
    """
    _group_id_description = """
    Exists if this transaction is part of a group. Denotes the parent’s transaction ID
    """
    _external_id_description = """
    User-defined external ID for any manually-entered or imported transaction.
    External ID cannot be accessed or changed for Plaid-imported transactions.
    External ID must be unique by asset_id. Max 75 characters.
    """
    _original_name_description = """
    The transactions original name before any payee name updates. For synced transactions,
    this is the raw original payee name from your bank.
    """
    _type_description = """
    (for synced investment transactions only) The transaction type as set by
    Plaid for investment transactions. Possible values include: buy, sell, cash,
    transfer and more
    """
    _subtype_description = """
    (for synced investment transactions only) The transaction type as set by Plaid
    for investment transactions. Possible values include: management fee, withdrawal,
    dividend, deposit and more
    """
    _fees_description = """
    (for synced investment transactions only) The fees as set by Plaid for investment
    transactions.
    """
    _price_description = """
    (for synced investment transactions only) The price as set by Plaid for investment
    transactions.
    """
    _quantity_description = """
    (for synced investment transactions only) The quantity as set by Plaid for investment
    transactions.
    """

    id: int = Field(description="Unique identifier for transaction")
    date: datetime.date = Field(description="Date of transaction in ISO 8601 format")
    payee: Optional[str] = Field(description=_payee_description)
    amount: float = Field(description=_amount_description)
    currency: Optional[str] = Field(max_length=3, description=_currency_description)
    notes: Optional[str] = Field(description=_notes_description)
    category_id: Optional[int] = Field(description=_category_description)
    asset_id: Optional[int] = Field(description=_asset_id_description)
    plaid_account_id: Optional[int] = Field(description=_plaid_account_id_description)
    status: Optional[str] = Field(description=_status_description)
    parent_id: Optional[int] = Field(description=_parent_id_description)
    is_group: Optional[bool] = Field(description=_is_group_description)
    group_id: Optional[int] = Field(description=_group_id_description)
    tags: Optional[List[TagsObject]] = Field(description="Array of Tag objects")
    external_id: Optional[str] = Field(
        max_length=75, description=_external_id_description
    )
    original_name: Optional[str] = Field(description=_original_name_description)
    type: Optional[str] = Field(description=_type_description)
    subtype: Optional[str] = Field(description=_subtype_description)
    fees: Optional[str] = Field(description=_fees_description)
    price: Optional[str] = Field(description=_price_description)
    quantity: Optional[str] = Field(description=_quantity_description)

    def get_update_object(self) -> TransactionUpdateObject:
        """
        Return a TransactionUpdateObject

        Return a TransactionUpdateObject to update an expense. Simply
        change one of the properties and perform an `update_transaction` with
        your Lunchable object.

        Returns
        -------
        TransactionUpdateObject
        """
        update_dict = self.dict()
        try:
            TransactionUpdateObject.StatusEnum(self.status)
        except ValueError:
            update_dict["status"] = None
        update_object = TransactionUpdateObject(**update_dict)
        if update_object.tags is not None:
            tags = [] if self.tags is None else self.tags
            update_object.tags = [tag.name for tag in tags]
        return update_object

    def get_insert_object(self) -> TransactionInsertObject:
        """
        Return a TransactionInsertObject

        Return a TransactionInsertObject to update an expense. Simply
        change some of the properties and perform an `insert_transactions` with
        your Lunchable object.

        Returns
        -------
        TransactionInsertObject
        """
        insert_dict = self.dict()
        try:
            TransactionInsertObject.StatusEnum(self.status)
        except ValueError:
            insert_dict["status"] = None
        insert_object = TransactionInsertObject(**insert_dict)
        if insert_object.tags is not None:
            tags = [] if self.tags is None else self.tags
            insert_object.tags = [tag.name for tag in tags]
        return insert_object


class _TransactionParamsGet(LunchableModel):
    """
    https://lunchmoney.dev/#get-all-transactions
    """

    tag_id: Optional[int]
    recurring_id: Optional[int]
    plaid_account_id: Optional[int]
    category_id: Optional[int]
    asset_id: Optional[int]
    group_id: Optional[int]
    is_group: Optional[bool]
    status: Optional[str]
    offset: Optional[int]
    limit: Optional[int]
    start_date: Optional[datetime.date]
    end_date: Optional[datetime.date]
    debit_as_negative: Optional[bool]


class _TransactionInsertParamsPost(LunchableModel):
    """
    https://lunchmoney.dev/#insert-transactions
    """

    transactions: List[TransactionInsertObject]
    apply_rules: bool = False
    skip_duplicates: bool = False
    check_for_recurring: bool = False
    debit_as_negative: bool = False
    skip_balance_update: bool = True


class _TransactionGroupParamsPost(LunchableModel):
    """
    https://lunchmoney.dev/#create-transaction-group
    """

    date: datetime.date
    payee: str
    category_id: Optional[int]
    notes: Optional[str]
    tags: Optional[List[int]]
    transactions: List[int]


class _TransactionUpdateParamsPut(LunchableModel):
    """
    https://lunchmoney.dev/#update-transaction
    """

    split: Optional[List[TransactionSplitObject]] = None
    transaction: Optional[TransactionUpdateObject] = None
    debit_as_negative: bool = False
    skip_balance_update: bool = True


class _TransactionsUnsplitPost(LunchableModel):
    """
    https://lunchmoney.dev/#unsplit-transactions
    """

    parent_ids: List[int]
    remove_parents: bool = False


class TransactionsClient(LunchMoneyAPIClient):
    """
    Lunch Money Transactions Interactions
    """

    def get_transactions(
        self,
        start_date: Optional[Union[datetime.date, datetime.datetime, str]] = None,
        end_date: Optional[Union[datetime.date, datetime.datetime, str]] = None,
        tag_id: Optional[int] = None,
        recurring_id: Optional[int] = None,
        plaid_account_id: Optional[int] = None,
        category_id: Optional[int] = None,
        asset_id: Optional[int] = None,
        group_id: Optional[int] = None,
        is_group: Optional[bool] = None,
        status: Optional[str] = None,
        offset: Optional[int] = None,
        limit: Optional[int] = None,
        debit_as_negative: Optional[bool] = None,
        params: Optional[dict] = None,
    ) -> List[TransactionObject]:
        """
        Get Transactions Using Criteria

        Use this to retrieve all transactions between a date range. Returns list of Transaction
        objects. If no query parameters are set, this will return transactions for the
        current calendar month. If either start_date or end_date are datetime.datetime objects,
        they will be reduced to dates. If a string is provided, it will be attempted to be parsed
        as YYYY-MM-DD format.

        Parameters
        ----------
        start_date: Optional[Union[datetime.date, datetime.datetime, str]]
            Denotes the beginning of the time period to fetch transactions for. Defaults
            to beginning of current month. Required if end_date exists. Format: YYYY-MM-DD.
        end_date: Optional[Union[datetime.date, datetime.datetime, str]]
            Denotes the end of the time period you'd like to get transactions for.
            Defaults to end of current month. Required if start_date exists.
        tag_id: Optional[int]
            Filter by tag. Only accepts IDs, not names.
        recurring_id: Optional[int]
            Filter by recurring expense
        plaid_account_id: Optional[int]
            Filter by Plaid account
        category_id: Optional[int]
            Filter by category. Will also match category groups.
        asset_id: Optional[int]
            Filter by asset
        group_id: Optional[int]
            Filter by group_id (if the transaction is part of a specific group)
        is_group: Optional[bool]
            Filter by group (returns transaction groups)
        status: Optional[str]
            Filter by status (Can be cleared or uncleared. For recurring
            transactions, use recurring)
        offset: Optional[int]
            Sets the offset for the records returned
        limit: Optional[int]
            Sets the maximum number of records to return. Note: The server will not
            respond with any indication that there are more records to be returned.
            Please check the response length to determine if you should make another
            call with an offset to fetch more transactions.
        debit_as_negative: Optional[bool]
            Pass in true if you’d like expenses to be returned as negative amounts and
            credits as positive amounts. Defaults to false.
        params: Optional[dict]
            Additional Query String Params

        Returns
        -------
        List[TransactionObject]
            A list of transactions

        Examples
        --------
        Retrieve a list of :class:`.TransactionObject` ::

            from lunchable import LunchMoney

            lunch = LunchMoney(access_token="xxxxxxx")
            transactions = lunch.get_transactions(start_date="2020-01-01",
                                                  end_date="2020-01-31")
        """
        search_params = _TransactionParamsGet(
            tag_id=tag_id,
            recurring_id=recurring_id,
            plaid_account_id=plaid_account_id,
            category_id=category_id,
            asset_id=asset_id,
            group_id=group_id,
            is_group=is_group,
            status=status,
            offset=offset,
            limit=limit,
            start_date=start_date,
            end_date=end_date,
            debit_as_negative=debit_as_negative,
        ).dict(exclude_none=True)
        search_params.update(params if params is not None else {})
        response_data = self._make_request(
            method=self.Methods.GET,
            url_path=APIConfig.LUNCHMONEY_TRANSACTIONS,
            params=search_params,
        )
        transactions = response_data[APIConfig.LUNCHMONEY_TRANSACTIONS]
        transaction_objects = [TransactionObject(**item) for item in transactions]
        return transaction_objects

    def get_transaction(self, transaction_id: int) -> TransactionObject:
        """
        Get a Transaction by ID

        Parameters
        ----------
        transaction_id: int
            Lunch Money Transaction ID

        Returns
        -------
        TransactionObject

        Examples
        --------
        Retrieve a single transaction by its ID ::

            from lunchable import LunchMoney

            lunch = LunchMoney(access_token="xxxxxxx")
            transaction = lunch.get_transaction(transaction_id=1234)

        The above code returns a :class:`.TransactionObject` with ID # 1234 (assuming
        it exists)
        """
        response_data = self._make_request(
            method=self.Methods.GET,
            url_path=[APIConfig.LUNCHMONEY_TRANSACTIONS, transaction_id],
        )
        return TransactionObject(**response_data)

    ListOrSingleTransactionUpdateObject = Union[
        TransactionUpdateObject, TransactionObject
    ]

    ListOrSingleTransactionInsertObject = Union[
        TransactionObject,
        TransactionInsertObject,
        List[TransactionObject],
        List[TransactionInsertObject],
        List[
            Union[TransactionObject, TransactionInsertObject],
        ],
    ]

    def update_transaction(
        self,
        transaction_id: int,
        transaction: ListOrSingleTransactionUpdateObject = None,
        split: Optional[List[TransactionSplitObject]] = None,
        debit_as_negative: bool = False,
        skip_balance_update: bool = True,
    ) -> Dict[str, Any]:
        """
        Update a Transaction

        Use this endpoint to update a single transaction. You may also use this
        to split an existing transaction. If a TransactionObject is provided it will be
        converted into a TransactionUpdateObject.

        PUT https://dev.lunchmoney.app/v1/transactions/:transaction_id

        Parameters
        ----------
        transaction_id: int
            Lunch Money Transaction ID
        transaction: ListOrSingleTransactionUpdateObject
            Object to update with
        split: Optional[List[TransactionSplitObject]]
            Defines the split of a transaction. You may not split an already-split
            transaction, recurring transaction, or group transaction.
        debit_as_negative: bool
            If true, will assume negative amount values denote expenses and
            positive amount values denote credits. Defaults to false.
        skip_balance_update: bool
            If false, will skip updating balance if an asset_id
            is present for any of the transactions.

        Returns
        -------
        Dict[str, Any]

        Examples
        --------
        Update a transaction with a :class:`.TransactionUpdateObject` ::

            from datetime import datetime

            from lunchable import LunchMoney, TransactionUpdateObject

            lunch = LunchMoney(access_token="xxxxxxx")
            transaction_note = f"Updated on {datetime.now()}"
            notes_update = TransactionUpdateObject(notes=transaction_note)
            response = lunch.update_transaction(transaction_id=1234,
                                                transaction=notes_update)

        Update a :class:`.TransactionObject` with itself ::

            from datetime import datetime, timedelta

            from lunchable import LunchMoney

            lunch = LunchMoney(access_token="xxxxxxx")
            transaction = lunch.get_transaction(transaction_id=1234)

            transaction.notes = f"Updated on {datetime.now()}"
            transaction.date = transaction.date + timedelta(days=1)
            response = lunch.update_transaction(transaction_id=transaction.id,
                                                transaction=transaction)
        """
        payload = _TransactionUpdateParamsPut(
            split=split,
            debit_as_negative=debit_as_negative,
            skip_balance_update=skip_balance_update,
        ).dict(exclude_none=True)
        if transaction is None and split is None:
            raise LunchMoneyError("You must update the transaction or provide a split")
        elif transaction is not None:
            if isinstance(transaction, TransactionObject):
                transaction = transaction.get_update_object()
            payload["transaction"] = transaction.dict(exclude_unset=True)
        response_data = self._make_request(
            method=self.Methods.PUT,
            url_path=[APIConfig.LUNCHMONEY_TRANSACTIONS, transaction_id],
            payload=payload,
        )
        return response_data

    def insert_transactions(
        self,
        transactions: ListOrSingleTransactionInsertObject,
        apply_rules: bool = False,
        skip_duplicates: bool = True,
        debit_as_negative: bool = False,
        check_for_recurring: bool = False,
        skip_balance_update: bool = True,
    ) -> List[int]:
        """
        Create One or Many Lunch Money Transactions

        Use this endpoint to insert many transactions at once. Also accepts
        a single transaction as well. If a TransactionObject is provided it will be
        converted into a TransactionInsertObject.

        https://lunchmoney.dev/#insert-transactions

        Parameters
        ----------
        transactions: ListOrSingleTransactionTypeObject
            Transactions to insert. Either a single TransactionInsertObject object or
            a list of them
        apply_rules: bool
            If true, will apply account’s existing rules to the inserted transactions.
            Defaults to false.
        skip_duplicates: bool
            If true, the system will automatically dedupe based on transaction date,
            payee and amount. Note that deduping by external_id will occur regardless
            of this flag.
        check_for_recurring: bool
            if true, will check new transactions for occurrences of new monthly expenses.
            Defaults to false.
        debit_as_negative: bool
            If true, will assume negative amount values denote expenses and
            positive amount values denote credits. Defaults to false.
        skip_balance_update: bool
            If false, will skip updating balance if an asset_id
            is present for any of the transactions.

        Returns
        -------
        List[int]

        Examples
        --------
        Create a new transaction with a :class:`.TransactionInsertObject` ::

            from lunchable import LunchMoney, TransactionInsertObject

            lunch = LunchMoney(access_token="xxxxxxx")

            new_transaction = TransactionInsertObject(payee="Example Restaurant",
                                                      amount=120.00,
                                                      notes="Saturday Dinner")
            new_transaction_ids = lunch.insert_transactions(transactions=new_transaction)
        """
        insert_objects = list()
        if not isinstance(transactions, list):
            transactions = [transactions]
        for item in transactions:
            if isinstance(item, TransactionObject):
                insert_objects.append(item.get_insert_object())
            elif isinstance(item, TransactionInsertObject):
                insert_objects.append(item)
            else:
                raise LunchMoneyError(
                    "Only TransactionObjects or TransactionInsertObjects are "
                    "supported by this function."
                )
        payload = _TransactionInsertParamsPost(
            transactions=insert_objects,
            apply_rules=apply_rules,
            skip_duplicates=skip_duplicates,
            check_for_recurring=check_for_recurring,
            debit_as_negative=debit_as_negative,
            skip_balance_update=skip_balance_update,
        ).dict(exclude_unset=True)
        response_data = self._make_request(
            method=self.Methods.POST,
            url_path=APIConfig.LUNCHMONEY_TRANSACTIONS,
            payload=payload,
        )
        ids: List[int] = response_data["ids"]
        return ids

    def insert_transaction_group(
        self,
        date: datetime.date,
        payee: str,
        transactions: List[int],
        category_id: Optional[int] = None,
        notes: Optional[str] = None,
        tags: Optional[List[int]] = None,
    ) -> int:
        """
        Create a Transaction Group of Two or More Transactions

        Returns the ID of the newly created transaction group

        Parameters
        ----------
        date: datetime.date
            Date for the grouped transaction
        payee: str
            Payee name for the grouped transaction
        category_id: Optional[int]
            Category for the grouped transaction
        notes: Optional[str]
            Notes for the grouped transaction
        tags: Optional[List[int]]
            Array of tag IDs for the grouped transaction
        transactions: Optional[List[int]]
            Array of transaction IDs to be part of the transaction group

        Returns
        -------
        int
        """
        if len(transactions) < 2:
            raise LunchMoneyError(
                "You must include 2 or more transactions " "in the Transaction Group"
            )
        transaction_params = _TransactionGroupParamsPost(
            date=date,
            payee=payee,
            category_id=category_id,
            notes=notes,
            tags=tags,
            transactions=transactions,
        ).dict(exclude_none=True)
        response_data = self._make_request(
            method=self.Methods.POST,
            url_path=[
                APIConfig.LUNCHMONEY_TRANSACTIONS,
                APIConfig.LUNCHMONEY_TRANSACTION_GROUPS,
            ],
            payload=transaction_params,
        )
        return response_data

    def remove_transaction_group(self, transaction_group_id: int) -> List[int]:
        """
        Delete a Transaction Group

        Use this method to delete a transaction group. The transactions within the
        group will not be removed.

        Returns the IDs of the transactions that were part of the deleted group

        https://lunchmoney.dev/#delete-transaction-group

        Parameters
        ----------
        transaction_group_id: int
            Transaction Group Identifier

        Returns
        -------
        List[int]
        """
        response_data = self._make_request(
            method=self.Methods.DELETE,
            url_path=[
                APIConfig.LUNCHMONEY_TRANSACTIONS,
                APIConfig.LUNCHMONEY_TRANSACTION_GROUPS,
                transaction_group_id,
            ],
        )
        return response_data["transactions"]

    def unsplit_transactions(
        self, parent_ids: List[int], remove_parents: bool = False
    ) -> List[int]:
        """
        Unsplit Transactions

        Use this endpoint to unsplit one or more transactions.

        Returns an array of IDs of deleted transactions

        https://lunchmoney.dev/#unsplit-transactions

        Parameters
        ----------
        parent_ids: List[int]
            Array of transaction IDs to unsplit. If one transaction is unsplittable,
            no transaction will be unsplit.
        remove_parents: bool
            If true, deletes the original parent transaction as well. Note,
            this is unreversable!

        Returns
        -------
        List[int]
        """
        response_data = self._make_request(
            method=self.Methods.POST,
            url_path=[APIConfig.LUNCHMONEY_TRANSACTIONS, "unsplit"],
            payload=_TransactionsUnsplitPost(
                parent_ids=parent_ids, remove_parents=remove_parents
            ).dict(exclude_none=True),
        )
        return response_data



"""
Lunchmoney CLI
"""

import json
import logging

import click
from pydantic.json import pydantic_encoder

import lunchable
from lunchable import LunchMoney
from lunchable.plugins.pushlunch import PushLunch

logger = logging.getLogger(__name__)


@click.group()
@click.version_option(lunchable.__version__)
def cli():
    """
    Interactions with Lunch Money via lunchable 🍱
    """
    pass


@cli.group()
def transactions():
    """
    Interact with Lunch Money transactions
    """


@cli.group()
def plugins():
    """
    Interact with Lunchable Plugins
    """


@transactions.command("get")
@click.option(
    "--start-date",
    default=None,
    help="Denotes the beginning of the time period to fetch transactions for. Defaults"
    "to beginning of current month. Required if end_date exists. "
    "Format: YYYY-MM-DD.",
)
@click.option(
    "--end-date",
    default=None,
    help="Denotes the end of the time period you'd like to get transactions for. "
    "Defaults to end of current month. Required if start_date exists."
    "Format: YYYY-MM-DD.",
)
@click.option(
    "--tag-id", default=None, help="Filter by tag. Only accepts IDs, not names."
)
@click.option("--recurring-id", default=None, help="Filter by recurring expense")
@click.option("--plaid-account-id", default=None, help="Filter by Plaid account")
@click.option(
    "--category-id",
    default=None,
    help="Filter by category. Will also match category groups.",
)
@click.option("--asset-id", default=None, help="Filter by asset")
@click.option(
    "--group-id",
    default=None,
    help="Filter by group_id (if the transaction is part of a specific group)",
)
@click.option(
    "--is-group", default=None, help="Filter by group (returns transaction groups)"
)
@click.option(
    "--status",
    default=None,
    help="Filter by status (Can be cleared or uncleared. For recurring "
    "transactions, use recurring)",
)
@click.option("--offset", default=None, help="Sets the offset for the records returned")
@click.option(
    "--limit",
    default=None,
    help="Sets the maximum number of records to return. Note: The server will not "
    "respond with any indication that there are more records to be returned. "
    "Please check the response length to determine if you should make another "
    "call with an offset to fetch more transactions.",
)
@click.option(
    "--debit-as-negative",
    default=None,
    help="Pass in true if you’d like expenses to be returned as negative amounts and "
    "credits as positive amounts. Defaults to false.",
)
def lunchmoney_transactions(**kwargs):
    """
    Retrieve Lunch Money Transactions
    """
    lunch = LunchMoney()
    transactions = lunch.get_transactions(**kwargs)
    click.echo(json.dumps(transactions, default=pydantic_encoder, indent=2))


@plugins.group()
def splitlunch():
    """
    Splitwise Plugin for lunchable, SplitLunch 💲🍱
    """
    pass


@splitlunch.command("expenses")
@click.option(
    "--limit", default=None, help="Limit the amount of Results. 0 returns everything."
)
@click.option("--offset", default=None, help="Number of expenses to be skipped")
@click.option("--limit", default=None, help="Number of expenses to be returned")
@click.option("--group-id", default=None, help="GroupID of the expenses")
@click.option("--friendship-id", default=None, help="FriendshipID of the expenses")
@click.option(
    "--dated-after",
    default=None,
    help="ISO 8601 Date time. Return expenses later that this date",
)
@click.option(
    "--dated-before",
    default=None,
    help="ISO 8601 Date time. Return expenses earlier than this date",
)
@click.option(
    "--updated-after",
    default=None,
    help="ISO 8601 Date time. Return expenses updated after this date",
)
@click.option(
    "--updated-before",
    default=None,
    help="ISO 8601 Date time. Return expenses updated before this date",
)
def splitlunch_expenses(**kwargs):
    """
    Retrieve Splitwise Expenses
    """
    from lunchable.plugins.splitlunch import SplitLunch

    splitlunch = SplitLunch()
    if set(kwargs.values()) == {None}:
        kwargs["limit"] = 5
    expenses = splitlunch.get_expenses(**kwargs)
    click.echo(json.dumps(expenses, default=pydantic_encoder, indent=2))


tag_transactions = click.option(
    "--tag-transactions",
    is_flag=True,
    help="Tag the resulting transactions with a `Splitwise` tag.",
)
financial_partner_id = click.option(
    "--financial-partner-id",
    default=None,
    help="Splitwise ID of your financial partner.",
)
financial_partner_email = click.option(
    "--financial-partner-email",
    default=None,
    help="Splitwise Email Address of your financial partner.",
)


@splitlunch.command("splitlunch")
@tag_transactions
def make_splitlunch(**kwargs):
    """
    Split all `SplitLunch` tagged transactions in half.

    One of these new splits will be recategorized to `Reimbursement`.
    """
    from lunchable.plugins.splitlunch import SplitLunch

    splitlunch = SplitLunch()
    results = splitlunch.make_splitlunch(**kwargs)
    click.echo(json.dumps(results, default=pydantic_encoder))


@splitlunch.command("splitlunch-import")
@tag_transactions
@financial_partner_id
@financial_partner_email
def make_splitlunch_import(**kwargs):
    """
    Import `SplitLunchImport` tagged transactions to Splitwise and Split them in Lunch Money

    Send a transaction to Splitwise and then split the original transaction in Lunch Money.
    One of these new splits will be recategorized to `Reimbursement`. Any tags will be
    reapplied.
    """
    from lunchable.plugins.splitlunch import SplitLunch

    financial_partner_id = kwargs.pop("financial_partner_id")
    financial_partner_email = kwargs.pop("financial_partner_email")
    splitlunch = SplitLunch(
        financial_partner_id=financial_partner_id,
        financial_partner_email=financial_partner_email,
    )
    results = splitlunch.make_splitlunch_import(**kwargs)
    click.echo(json.dumps(results, default=pydantic_encoder, indent=2))


@splitlunch.command("splitlunch-direct-import")
@tag_transactions
@financial_partner_id
@financial_partner_email
def make_splitlunch_direct_import(**kwargs):
    """
    Import `SplitLunchDirectImport` tagged transactions to Splitwise and Split them in Lunch Money

    Send a transaction to Splitwise and then split the original transaction in Lunch Money.
    One of these new splits will be recategorized to `Reimbursement`. Any tags will be
    reapplied.
    """
    from lunchable.plugins.splitlunch import SplitLunch

    financial_partner_id = kwargs.pop("financial_partner_id")
    financial_partner_email = kwargs.pop("financial_partner_email")
    splitlunch = SplitLunch(
        financial_partner_id=financial_partner_id,
        financial_partner_email=financial_partner_email,
    )
    results = splitlunch.make_splitlunch_direct_import(**kwargs)
    click.echo(json.dumps(results, default=pydantic_encoder))


@splitlunch.command("update-balance")
def update_splitwise_balance(**kwargs):
    """
    Update the Splitwise Asset Balance
    """
    from lunchable.plugins.splitlunch import SplitLunch

    splitlunch = SplitLunch()
    updated_asset = splitlunch.update_splitwise_balance()
    click.echo(json.dumps(updated_asset, default=pydantic_encoder, indent=2))


@splitlunch.command("refresh")
def refresh_splitwise_transactions(**kwargs):
    """
    Import New Splitwise Transactions to Lunch Money and

    This function gets all transactions from Splitwise, all transactions from
    your Lunch Money Splitwise account and compares the two. This also updates
    the account balance.
    """
    from lunchable.plugins.splitlunch import SplitLunch

    splitlunch = SplitLunch()
    response = splitlunch.refresh_splitwise_transactions()
    click.echo(json.dumps(response, default=pydantic_encoder, indent=2))


@plugins.group()
def pushlunch():
    """
    Push Notifications for Lunch Money: PushLunch 📲
    """
    pass


@pushlunch.command("notify")
@click.option(
    "--continuous",
    is_flag=True,
    help="Whether to continuously check for more uncleared transactions, "
    "waiting a fixed amount in between checks.",
)
@click.option(
    "--interval",
    default=None,
    help="Sleep Interval in Between Tries - only applies if `continuous` is set. "
    "Defaults to 60 (minutes). Cannot be less than 5 (minutes)",
)
@click.option(
    "--user-key",
    default=None,
    help="Pushover User Key. Defaults to `PUSHOVER_USER_KEY` env var",
)
def notify(continuous: bool, interval: int, user_key: str):
    """
    Send a Notification for each Uncleared Transaction
    """
    push = PushLunch(user_key=user_key)
    if interval is not None:
        interval = int(interval)
    if continuous is not None:
        logging.basicConfig(
            level=logging.INFO, format="%(asctime)s [%(levelname)8s]: %(message)s"
        )
    push.notify_uncleared_transactions(continuous=continuous, interval=interval)



"""
Lunch Money - Recurring Expenses

https://lunchmoney.dev/#recurring-expenses
"""

import datetime
import logging
from typing import List, Optional

from pydantic import Field

from lunchable._config import APIConfig
from lunchable.models._base import LunchableModel
from lunchable.models._core import LunchMoneyAPIClient

logger = logging.getLogger(__name__)


class RecurringExpensesObject(LunchableModel):
    """
    Recurring Expenses Object

    https://lunchmoney.dev/#recurring-expenses-object
    """

    _id_description = "Unique identifier for recurring expense"
    _start_date_description = """
    Denotes when recurring expense starts occurring in ISO 8601 format.
    If null, then this recurring expense will show up for all time
    before end_date
    """
    _end_date_description = """
    Denotes when recurring expense stops occurring in ISO 8601 format.
    If null, then this recurring expense has no set end date and will
    show up for all months after start_date
    """
    _cadence_description = """
    One of: [monthly, twice a month, once a week, every 3 months, every 4 months,
    twice a year, yearly]
    """
    _amount_description = (
        "Amount of the recurring expense in numeric format to 4 decimal places"
    )
    _currency_description = """
    Three-letter lowercase currency code for the recurring expense in ISO 4217 format
    """
    _description_description = """
    If any, represents the user-entered description of the recurring expense
    """
    _billing_date_description = """
    Expected billing date for this recurring expense for this month in ISO 8601 format
    """
    _type_description = """"
    This can be one of two values: cleared (The recurring expense has been reviewed
    by the user), suggested (The recurring expense is suggested by the system;
    the user has yet to review/clear it)
    """
    _original_name_description = """
    If any, represents the original name of the recurring expense as
    denoted by the transaction that triggered its creation
    """
    _source_description = """
    This can be one of three values: manual (User created this recurring expense
    manually from the Recurring Expenses page), transaction (User created this by
    converting a transaction from the Transactions page), system (Recurring expense
    was created by the system on transaction import). Some older recurring expenses
    may not have a source.
    """
    _plaid_account_id_description = """
    If any, denotes the plaid account associated with the creation of this "
    recurring expense (see Plaid Accounts)"
    """
    _asset_id_description = """
    If any, denotes the manually-managed account (i.e. asset) associated with the
    creation of this recurring expense (see Assets)
    """
    _transaction_id_description = """
    If any, denotes the unique identifier for the associated transaction matching
    this recurring expense for the current time period
    """
    _category_id_description = """
    If any, denotes the unique identifier for the associated category to this recurring expense
    """

    id: int = Field(description=_id_description)
    start_date: Optional[datetime.date] = Field(description=_start_date_description)
    end_date: Optional[datetime.date] = Field(description=_end_date_description)
    cadence: str = Field(description=_cadence_description)
    payee: str = Field(description="Payee of the recurring expense")
    amount: float = Field(description=_amount_description)
    currency: str = Field(max_length=3, description=_currency_description)
    description: Optional[str] = Field(description=_description_description)
    billing_date: datetime.date = Field(description=_billing_date_description)
    type: str = Field(description=_type_description)
    original_name: Optional[str] = Field(description=_original_name_description)
    source: str = Field(description=_source_description)
    plaid_account_id: Optional[int] = Field(description=_plaid_account_id_description)
    asset_id: Optional[int] = Field(description=_asset_id_description)
    transaction_id: Optional[int] = Field(description=_transaction_id_description)
    category_id: Optional[int] = Field(description=_category_id_description)


class RecurringExpenseParamsGet(LunchableModel):
    """
    https://lunchmoney.dev/#get-recurring-expenses
    """

    start_date: datetime.date
    debit_as_negative: bool


class RecurringExpensesClient(LunchMoneyAPIClient):
    """
    Lunch Money Recurring Expenses Interactions
    """

    def get_recurring_expenses(
        self,
        start_date: Optional[datetime.date] = None,
        debit_as_negative: bool = False,
    ) -> List[RecurringExpensesObject]:
        """
        Get Recurring Expenses

        Retrieve a list of recurring expenses to expect for a specified period.

        Every month, a different set of recurring expenses is expected. This is because recurring
        expenses can be once a year, twice a year, every 4 months, etc.

        If a recurring expense is listed as “twice a month”, then that recurring expense will be
        returned twice, each with a different billing date based on when the system believes that
        recurring expense transaction is to be expected. If the recurring expense is listed as
        “once a week”, then that recurring expense will be returned in this list as many times as
        there are weeks for the specified month.

        In the same vein, if a recurring expense that began last month is set to “Every 3
        months”, then that recurring expense will not show up in the results for this month.

        Parameters
        ----------
        start_date : Optional[datetime.date]
            Date to search. By default will return the first day of the current month
        debit_as_negative: bool
            Pass in true if you’d like expenses to be returned as negative amounts and credits as
            positive amounts.

        Returns
        -------
        List[RecurringExpensesObject]
        """
        if start_date is None:
            start_date = datetime.datetime.now().date().replace(day=1)
        params = RecurringExpenseParamsGet(
            start_date=start_date, debit_as_negative=debit_as_negative
        ).dict()
        response_data = self._make_request(
            method="GET",
            url_path=[APIConfig.LUNCH_MONEY_RECURRING_EXPENSES],
            params=params,
        )
        recurring_expenses = response_data.get(APIConfig.LUNCH_MONEY_RECURRING_EXPENSES)
        recurring_expenses_objects = [
            RecurringExpensesObject(**item) for item in recurring_expenses
        ]
        logger.debug(
            "%s RecurringExpensesObjects retrieved", len(recurring_expenses_objects)
        )
        return recurring_expenses_objects



"""
API Configuration Helper
"""

import logging
from os import getenv
from typing import Dict, List, Optional, Union
from urllib import parse

from lunchable.exceptions import EnvironmentVariableError, LunchMoneyError

logger = logging.getLogger(__name__)


class APIConfig:
    """
    Configuration Helper Class for Connecting to the Lunchmoney API
    """

    LUNCHMONEY_SCHEME: str = "https"
    LUNCHMONEY_NETLOC: str = "dev.lunchmoney.app"
    LUNCHMONEY_API_PATH: str = "v1"

    LUNCHMONEY_TRANSACTIONS: str = "transactions"
    LUNCHMONEY_TRANSACTION_GROUPS: str = "group"
    LUNCHMONEY_PLAID_ACCOUNTS: str = "plaid_accounts"
    LUNCH_MONEY_RECURRING_EXPENSES: str = "recurring_expenses"
    LUNCHMONEY_BUDGET: str = "budgets"
    LUNCHMONEY_ASSETS: str = "assets"
    LUNCHMONEY_CATEGORIES: str = "categories"
    LUNCHMONEY_TAGS: str = "tags"
    LUNCHMONEY_CRYPTO: str = "crypto"
    LUNCHMONEY_CRYPTO_MANUAL: str = "manual"

    LUNCHMONEY_CONTENT_TYPE_HEADERS: Dict[str, str] = {
        "Content-Type": "application/json"
    }

    _access_token_environment_variable = "LUNCHMONEY_ACCESS_TOKEN"

    @staticmethod
    def get_access_token(access_token: Optional[str] = None) -> str:
        """
        Method for Resolving Access Tokens: Hardcoded -> .env File -> Env Var

        Parameters
        ----------
        access_token: Optional[str]
            Lunchmoney Developer API Access Token

        Returns
        -------
        str
        """
        if access_token is None:
            logger.info(
                "Loading Lunch Money Developer API Access token from environment"
            )
            access_token = getenv(APIConfig._access_token_environment_variable, None)
        if access_token is None:
            access_token_error_message = (
                "You must provide a Lunch Money Developer API Access Token directly or set your "
                f"{APIConfig._access_token_environment_variable} environment variable."
            )
            raise EnvironmentVariableError(access_token_error_message)
        return access_token

    @staticmethod
    def get_header(access_token: Optional[str] = None) -> Dict[str, str]:
        """
        Get the header dict to pass to requests

        Parameters
        ----------
        access_token: Optional[str]
            Lunchmoney Developer API Access Token

        Returns
        -------
        Dict[str, str]
        """
        access_token = APIConfig.get_access_token(access_token=access_token)
        auth_header = {"Authorization": f"Bearer {access_token}"}
        auth_header.update(APIConfig.LUNCHMONEY_CONTENT_TYPE_HEADERS)
        return auth_header

    @staticmethod
    def make_url(url_path: Union[List[Union[str, int]], str, int]):
        """
        Make a Lunch Money API URL using path parts

        Parameters
        ----------
        url_path: Union[List[Union[str, int]], str, int]
            API Components, if a list join these sequentially

        Returns
        -------
        str
        """
        if isinstance(url_path, str):
            url_path = [url_path]
        if not isinstance(url_path, List):
            raise LunchMoneyError(
                "You must provide a string or list of strings to construct a URL"
            )
        path_set = [
            str(item).lower()
            for item in url_path
            if str(item).lower() != APIConfig.LUNCHMONEY_API_PATH
        ]
        url = APIConfig._generate_url(
            scheme=APIConfig.LUNCHMONEY_SCHEME,
            netloc=APIConfig.LUNCHMONEY_NETLOC,
            path="/".join([APIConfig.LUNCHMONEY_API_PATH, *path_set]),
        )
        return url

    @classmethod
    def _generate_url(
        cls,
        scheme: str,
        netloc: str,
        path: str = "",
        params: str = "",
        query: str = "",
        fragment: str = "",
    ):
        """
        Build a URL

        Parameters
        ----------
        scheme: str
            URL scheme specifier
        netloc: str
            Network location part
        path: str
            Hierarchical path
        params: str
            Parameters for last path element
        query: str
            Query component
        fragment: str
            Fragment identifier
        Returns
        -------
        url: str
            Compiled URL
        """
        url_components = dict(
            scheme=scheme,
            netloc=netloc,
            path=path,
            params=params,
            query=query,
            fragment=fragment,
        )
        return parse.urlunparse(components=tuple(url_components.values()))



"""
Lunchmoney SDK Core
"""

import datetime
import json
import logging
from json import loads
from typing import Any, List, Optional, Union

import requests

from lunchable._config import APIConfig
from lunchable.exceptions import LunchMoneyHTTPError

logger = logging.getLogger(__name__)


class _Methods:
    """
    Namespace for Request Methods
    """

    GET = "GET"
    POST = "POST"
    PUT = "PUT"
    DELETE = "DELETE"


class LunchMoneyAPIClient:
    """
    Core API Client Class
    """

    class Methods:
        """
        HTTP Request Method Enumerations: GET, OPTIONS, HEAD, POST, PUT, PATCH, or DELETE
        """

        # This Helper Namespace Organizes and Tracks HTTP Requests by Method
        GET = "GET"
        OPTIONS = "OPTIONS"
        HEAD = "HEAD"
        POST = "POST"
        PUT = "PUT"
        PATCH = "PATCH"
        DELETE = "DELETE"

    def __init__(self, access_token: Optional[str] = None):
        """
        Initialize a Lunch Money object with an Access Token.

        Tries to inherit from the Environment if one isn't provided

        Parameters
        ----------
        access_token: Optional[str]
            Lunchmoney Developer API Access Token
        """
        self.access_token = APIConfig.get_access_token(access_token=access_token)
        api_headers = APIConfig.get_header(access_token=self.access_token)
        default_headers = requests.sessions.default_headers()
        updated_headers = dict(**default_headers, **api_headers)
        typed_headers = requests.models.CaseInsensitiveDict(updated_headers)
        self.lunch_money_session = requests.Session()
        self.lunch_money_session.headers = typed_headers

    def __repr__(self) -> str:
        """
        String Representation

        Returns
        -------
        str
        """
        return "<LunchMoney: requests.Session>"

    @staticmethod
    def _serializer(obj):
        if isinstance(obj, datetime.datetime):
            return obj.isoformat()
        elif isinstance(obj, datetime.date):
            return obj.isoformat()
        else:
            return obj

    def _make_request(
        self,
        method: str,
        url_path: Union[List[Union[str, int]], str, int],
        params: Optional[dict] = None,
        payload: Optional[Any] = None,
        **kwargs
    ) -> Any:
        """
        Make a Request to the API

        Parameters
        ----------
        method: str
            requests method: GET, OPTIONS, HEAD, POST, PUT,
            PATCH, or DELETE
        url_path: Union[List[Union[str, int]], str, int]
            API Components, if a list join these sequentially
        params: Optional[dict]
            Params to pass
        payload: Optional[Any]
            data to pass

        Returns
        -------
        Any
        """
        url = APIConfig.make_url(url_path=url_path)
        if payload is not None:
            payload = json.dumps(payload, default=LunchMoneyAPIClient._serializer)
        try:
            response = self.lunch_money_session.request(
                method=method, url=url, params=params, data=payload, **kwargs
            )
            response.raise_for_status()
        except requests.exceptions.HTTPError as he:
            logger.exception(he)
            # noinspection PyUnboundLocalVariable
            logger.error(response.text)
            raise LunchMoneyHTTPError(he)
        returned_data = loads(response.content)
        if isinstance(returned_data, dict) and any(
            ["error" in returned_data.keys(), "errors" in returned_data.keys()]
        ):
            try:
                errors = returned_data["error"]
            except KeyError:
                errors = returned_data["errors"]
            logger.exception(errors)
            raise LunchMoneyHTTPError(errors)
        return loads(response.content)

    def make_http_request(
        self,
        method: str,
        url: str,
        params: Optional[Any] = None,
        data: Optional[Any] = None,
        json: Optional[Any] = None,
        **kwargs
    ) -> requests.Response:
        """
        Make a HTTP Request

        This is a simple method :class:`.LunchMoney` exposes to make HTTP requests. It
        has the benefit of using an existing `requests.Session` as well as as out of the box
        auth headers that are used to connect to the Lunch Money Developer API.

        Parameters
        ----------
        method: str
            requests method: GET, OPTIONS, HEAD, POST, PUT,
            PATCH, or DELETE
        url: str
            URL for the new Request object.
        params: Optional[Any]
            Dictionary, list of tuples or bytes to send in the query
            string for the Request.
        data: Optional[Any]
            Dictionary, list of tuples, bytes, or file-like object to send
            in the body of the Request.
        json: Optional[Any]
            A JSON serializable Python object to send in the body of the Request.

        Returns
        -------
        Any

        Examples
        --------
        A recent use of this method was to delete a Tag (which isn't available via the
        Developer API yet) ::

            import lunchable

            lunch = lunchable.LunchMoney()

            # Get All the Tags
            all_tags = lunch.get_tags()
            # Get All The Null Tags (a list of 1 or zero)
            null_tags = [tag for tag in all_tags if tag.name in [None, ""]]

            # Create a Cookie dictionary from a browser session
            cookies = {"cookie_keys": "cookie_values"}

            for null_tag in null_tags:
                # use the requests.session embedded in the class to make a request with cookies
                response = lunch.make_request(
                    method="DELETE",
                    url=f"https://api.lunchmoney.app/tags/{null_tag.id}",
                    cookies=cookies)
                # raise an error for 4XX responses
                response.raise_for_status()
        """
        response = self.lunch_money_session.request(
            method=method, url=url, params=params, data=data, json=json, **kwargs
        )
        return response



"""
Splitwise Plugin for Lunchmoney
"""

from .exceptions import SplitLunchError
from .lunchmoney_splitwise import SplitLunch
from .models import SplitLunchExpense

__all__ = ["SplitLunch", "SplitLunchError", "SplitLunchExpense"]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['lunchable',
 'lunchable._config',
 'lunchable.models',
 'lunchable.plugins',
 'lunchable.plugins.pushlunch',
 'lunchable.plugins.splitlunch']

package_data = \
{'': ['*']}

install_requires = \
['click>=8.0.1', 'pydantic>=1.2.0', 'requests>=2,<3']

extras_require = \
{'dev': ['pytest>=5.2',
         'flake8>=3.0.0',
         'coverage[toml]>=6.2',
         'mypy>=0.971,<0.972',
         'tox>=3.25.1,<4.0.0'],
 'splitlunch': ['splitwise>=2.3.0', 'python-dateutil>=2.8.2,<3.0.0']}

entry_points = \
{'console_scripts': ['lunchable = lunchable._cli:cli']}

setup_kwargs = {
    'name': 'lunchable',
    'version': '0.1.19',
    'description': 'A simple Python SDK around the Lunch Money Developer API',
    'long_description': '# lunchable\n\n<div align="center">\n<a href="https://github.com/juftin/lunchable">\n  <img src=https://i.imgur.com/FyKDsG3.png\n    width="400" alt="lunchable">\n</a>\n</div>\n\n[![Lunchable Version](https://img.shields.io/pypi/v/lunchable?color=blue&label=lunchable)](https://github.com/juftin/lunchable)\n[![PyPI](https://img.shields.io/pypi/pyversions/lunchable)](https://pypi.python.org/pypi/lunchable/)\n[![Docker Image Version](https://img.shields.io/docker/v/juftin/lunchable?color=blue&label=docker&logo=docker)](https://hub.docker.com/r/juftin/lunchable)\n[![Testing Status](https://github.com/juftin/lunchable/actions/workflows/tests.yaml/badge.svg?branch=main)](https://github.com/juftin/lunchable/actions/workflows/tests.yaml?query=branch%3Amain)\n[![GitHub License](https://img.shields.io/github/license/juftin/lunchable?color=blue&label=License)](https://github.com/juftin/lunchable/blob/main/LICENSE)\n\n**lunchable** is a Python Client for the [Lunch Money Developer API](https://lunchmoney.dev). It\'s\nbuilt on top of [pydantic](https://github.com/samuelcolvin/pydantic), it offers an *intuitive* API,\na *simple* CLI, complete coverage of all endpoints, and *plugins* to other external services.\n\n### Installation\n\n```shell\npip install lunchable\n```\n\n### Usage\n\n```python\nfrom typing import Any, Dict, List\n\nfrom lunchable import LunchMoney\nfrom lunchable.models import TransactionObject\n\nlunch = LunchMoney(access_token="xxxxxxxxxxx")\ntransactions: List[TransactionObject] = lunch.get_transactions()\n\nfirst_transaction: TransactionObject = transactions[0]\ntransaction_as_dict: Dict[str, Any] = first_transaction.dict()\n```\n\n```shell\nexport LUNCHMONEY_ACCESS_TOKEN="xxxxxxxxxxx"\nlunchable transactions get --limit 5\n```\n\n### Check out the [**Docs**](https://juftin.com/lunchable/)\n### Looking to contribute? See the [Contributing Guide](docs/source/contributing.md)\n### See the [Changelog](https://github.com/juftin/lunchable/releases)\n\n\n--------------\n--------------\n\n<br/>\n\n[<p align="center" ><img src="https://raw.githubusercontent.com/juftin/juftin/main/static/juftin.png" width="60" height="60"  alt="juftin logo"> </p>](https://github.com/juftin)\n',
    'author': 'Justin Flannery',
    'author_email': 'juftin@juftin.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/juftin/lunchable',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



"""
Lunch Money - Tags

https://lunchmoney.dev/#tags
"""

import logging
from typing import List, Optional

from pydantic import Field

from lunchable._config import APIConfig
from lunchable.models._base import LunchableModel
from lunchable.models._core import LunchMoneyAPIClient

logger = logging.getLogger(__name__)


class TagsObject(LunchableModel):
    """
    Lunchmoney Tags object

    https://lunchmoney.dev/#tags-object
    """

    id: int = Field(description="Unique identifier for tag")
    name: str = Field(description="User-defined name of tag", min_length=1)
    description: Optional[str] = Field(description="User-defined description of tag")


class TagsClient(LunchMoneyAPIClient):
    """
    Lunch Money Tag Interactions
    """

    def get_tags(self) -> List[TagsObject]:
        """
        Get Spending Tags

        Use this endpoint to get a list of all tags associated with the
        user's account.

        https://lunchmoney.dev/#get-all-tags

        Returns
        -------
        List[TagsObject]
        """
        response_data = self._make_request(
            method=self.Methods.GET, url_path=APIConfig.LUNCHMONEY_TAGS
        )
        tag_objects = [TagsObject(**item) for item in response_data]
        return tag_objects



"""
Lunch Money - Budgets

https://lunchmoney.dev/#budget
"""

import datetime
import logging
from typing import Any, Dict, List, Optional

from pydantic import Field

from lunchable._config import APIConfig
from lunchable.models._base import LunchableModel
from lunchable.models._core import LunchMoneyAPIClient

logger = logging.getLogger(__name__)


class BudgetDataObject(LunchableModel):
    """
    Data Object within a Budget
    """

    budget_amount: Optional[float] = Field()
    budget_currency: Optional[str] = Field()
    budget_to_base: Optional[float] = Field()
    spending_to_base: float = Field(default=0.00)
    num_transactions: int = Field(default=0)


class BudgetConfigObject(LunchableModel):
    """
    Budget Configuration Object
    """

    config_id: int
    cadence: str
    amount: Optional[float]
    currency: Optional[str]
    to_base: Optional[float]
    auto_suggest: str


class BudgetObject(LunchableModel):
    """
    Monthly Budget Per Category Object

    https://lunchmoney.dev/#budget-object
    """

    _category_group_name_description = "Name of the category group, if applicable"
    _is_income_description = """
    If true, this category is an income category (category properties
    are set in the app via the Categories page)
    """
    _exclude_from_budget_description = """
    If true, this category is excluded from budget (category
    properties are set in the app via the Categories page)
    """
    _exclude_from_totals_description = """
    If true, this category is excluded from totals (category
    properties are set in the app via the Categories page)
    """
    _data_description = """
    For each month with budget or category spending data, there is a data object with the key
    set to the month in format YYYY-MM-DD. For properties, see Data object below.
    """
    _config_description = """
    Object representing the category's budget suggestion configuration
    """

    category_name: str = Field(description="Name of the category")
    category_id: Optional[int] = Field(description="Unique identifier for category")
    category_group_name: Optional[str] = Field(
        description=_category_group_name_description
    )
    group_id: Optional[int] = Field(description="Unique identifier for category group")
    is_group: Optional[bool] = Field(description="If true, this category is a group")
    is_income: bool = Field(description=_is_income_description)
    exclude_from_budget: bool = Field(description=_exclude_from_budget_description)
    exclude_from_totals: bool = Field(description=_exclude_from_totals_description)
    data: Dict[datetime.date, BudgetDataObject] = Field(description=_data_description)
    config: Optional[BudgetConfigObject] = Field(description=_config_description)


class BudgetParamsGet(LunchableModel):
    """
    https://lunchmoney.dev/#get-budget-summary
    """

    start_date: datetime.date
    end_date: datetime.date


class BudgetParamsPut(LunchableModel):
    """
    https://lunchmoney.dev/#upsert-budget
    """

    start_date: datetime.date
    category_id: int
    amount: float
    currency: Optional[str]


class BudgetParamsRemove(LunchableModel):
    """
    https://lunchmoney.dev/#remove-budget
    """

    start_date: datetime.date
    category_id: int


class BudgetsClient(LunchMoneyAPIClient):
    """
    Lunch Money Budget Interactions
    """

    def get_budgets(
        self, start_date: datetime.date, end_date: datetime.date
    ) -> List[BudgetObject]:
        """
        Get Monthly Budgets

        Get full details on the budgets for all categories between a certain time
        period. The budgeted and spending amounts will be an aggregate across this
        time period. (https://lunchmoney.dev/#plaid-accounts-object)

        Returns
        -------
        List[BudgetObject]
        """
        params = BudgetParamsGet(start_date=start_date, end_date=end_date).dict()
        response_data = self._make_request(
            method="GET", url_path=[APIConfig.LUNCHMONEY_BUDGET], params=params
        )
        budget_objects = [BudgetObject(**item) for item in response_data]
        return budget_objects

    def upsert_budget(
        self,
        start_date: datetime.date,
        category_id: int,
        amount: float,
        currency: Optional[str] = None,
    ) -> Optional[Dict[str, Any]]:
        """
        Upsert a Budget for a Category and Date

        Use this endpoint to update an existing budget or insert a new budget for
        a particular category and date.

        Note: Lunch Money currently only supports monthly budgets, so your date must
        always be the start of a month (eg. 2021-04-01)

        If this is a sub-category, the response will include the updated category
        group's budget. This is because setting a sub-category may also update
        the category group's overall budget.

        https://lunchmoney.dev/#upsert-budget

        Parameters
        ----------
        start_date : date
            Start date for the budget period. Lunch Money currently only supports monthly budgets,
            so your date must always be the start of a month (eg. 2021-04-01)
        category_id: int
            Unique identifier for the category
        amount: float
            Amount for budget
        currency: Optional[str]
            Currency for the budgeted amount (optional). If empty, will default to your primary
            currency

        Returns
        -------
        Optional[Dict[str, Any]]
        """
        body = BudgetParamsPut(
            start_date=start_date,
            category_id=category_id,
            amount=amount,
            currency=currency,
        ).dict(exclude_none=True)
        response_data = self._make_request(
            method="PUT", url_path=[APIConfig.LUNCHMONEY_BUDGET], payload=body
        )

        return response_data["category_group"]

    def remove_budget(self, start_date: datetime.date, category_id: int) -> bool:
        """
        Unset an Existing Budget for a Particular Category in a Particular Month

        Parameters
        ----------
        start_date : date
            Start date for the budget period. Lunch Money currently only supports monthly budgets,
            so your date must always be the start of a month (eg. 2021-04-01)
        category_id: int
            Unique identifier for the category

        Returns
        -------
        bool
        """
        params = BudgetParamsRemove(
            start_date=start_date, category_id=category_id
        ).dict()
        response_data = self._make_request(
            method="DELETE", url_path=[APIConfig.LUNCHMONEY_BUDGET], params=params
        )
        return response_data



"""
Lunch Money - Assets

https://lunchmoney.dev/#assets
"""

import datetime
import logging
from typing import List, Optional

from pydantic import Field, validator

from lunchable._config import APIConfig
from lunchable.models._base import LunchableModel
from lunchable.models._core import LunchMoneyAPIClient

logger = logging.getLogger(__name__)


class AssetsObject(LunchableModel):
    """
    Manually Managed Asset Objects

    Assets in Lunch Money are similar to `plaid-accounts` except that they are manually managed.

    https://lunchmoney.dev/#assets-object
    """

    _type_name_description = """
    Primary type of the asset. Must be one of: [employee compensation, cash, vehicle, loan,
    cryptocurrency, investment, other, credit, real estate]
    """
    _subtype_name_description = """
    Optional asset subtype. Examples include: [retirement, checking, savings, prepaid credit card]
    """
    _balance_description = (
        "Current balance of the asset in numeric format to 4 decimal places"
    )
    _balance_as_of_description = """
    Date/time the balance was last updated in ISO 8601 extended format
    """
    _closed_on_description = "The date this asset was closed (optional)"
    _currency_description = (
        "Three-letter lowercase currency code of the balance in ISO 4217 format"
    )
    _created_at_description = (
        "Date/time the asset was created in ISO 8601 extended format"
    )
    _exclude_transactions_description = (
        "If true, this asset will not show up as an "
        "option for assignment when creating "
        "transactions manually"
    )

    id: int = Field(description="Unique identifier for asset")
    type_name: str = Field(description=_type_name_description)
    subtype_name: Optional[str] = Field(description=_subtype_name_description)
    name: str = Field(description="Name of the asset")
    display_name: Optional[str] = Field(
        description="Display name of the asset (as set by user)"
    )
    balance: float = Field(description=_balance_description)
    balance_as_of: datetime.datetime = Field(description=_balance_as_of_description)
    closed_on: Optional[datetime.date] = Field(description=_closed_on_description)
    currency: str = Field(description=_currency_description)
    institution_name: Optional[str] = Field(
        description="Name of institution holding the asset"
    )
    exclude_transactions: bool = Field(
        default=False, description=_exclude_transactions_description
    )
    created_at: datetime.datetime = Field(description=_created_at_description)


class _AssetsParamsPut(LunchableModel):
    """
    https://lunchmoney.dev/#update-asset
    """

    type_name: Optional[str]
    subtype_name: Optional[str]
    name: Optional[str]
    balance: Optional[str]
    balance_as_of: Optional[datetime.datetime]
    currency: Optional[str]
    institution_name: Optional[str]

    @classmethod
    @validator("balance", pre=True)
    def result_check(cls, x):
        """
        Check a result
        """
        return round(x, 2)


class _AssetsParamsPost(LunchableModel):
    """
    https://lunchmoney.dev/#create-asset
    """

    type_name: str
    subtype_name: Optional[str]
    name: str
    display_name: Optional[str]
    balance: float
    balance_as_of: Optional[datetime.datetime]
    currency: Optional[str]
    institution_name: Optional[str]
    closed_on: Optional[datetime.date]
    exclude_transactions: bool = False

    @classmethod
    @validator("balance", pre=True)
    def result_check(cls, x):
        """
        Check a result
        """
        return round(x, 2)


class AssetsClient(LunchMoneyAPIClient):
    """
    Lunch Money Assets Interactions
    """

    def get_assets(self) -> List[AssetsObject]:
        """
        Get Manually Managed Assets

        Get a list of all manually-managed assets associated with the user's account.

        (https://lunchmoney.dev/#assets-object)

        Returns
        -------
        List[AssetsObject]
        """
        response_data = self._make_request(
            method=self.Methods.GET, url_path=[APIConfig.LUNCHMONEY_ASSETS]
        )
        assets = response_data.get(APIConfig.LUNCHMONEY_ASSETS)
        asset_objects = [AssetsObject(**item) for item in assets]
        return asset_objects

    def update_asset(
        self,
        asset_id: int,
        type_name: Optional[str] = None,
        subtype_name: Optional[str] = None,
        name: Optional[str] = None,
        balance: Optional[float] = None,
        balance_as_of: Optional[datetime.datetime] = None,
        currency: Optional[str] = None,
        institution_name: Optional[str] = None,
    ) -> AssetsObject:
        """
        Update a Single Asset

        Parameters
        ----------
        asset_id: int
            Asset Identifier
        type_name: Optional[str]
            Must be one of: cash, credit, investment, other, real estate, loan, vehicle,
            cryptocurrency, employee compensation
        subtype_name: Optional[str]
            Max 25 characters
        name: Optional[str]
            Max 45 characters
        balance: Optional[float]
            Numeric value of the current balance of the account. Do not include any special
            characters aside from a decimal point!
        balance_as_of: Optional[datetime.datetime]
            Has no effect if balance is not defined. If balance is defined, but balance_as_of
            is not supplied or is invalid, current timestamp will be used.
        currency: Optional[str]
            Three-letter lowercase currency in ISO 4217 format. The code sent must exist in
            our database. Defaults to asset's currency.
        institution_name: Optional[str]
            Max 50 characters

        Returns
        -------
        AssetsObject
        """
        payload = _AssetsParamsPut(
            type_name=type_name,
            subtype_name=subtype_name,
            name=name,
            balance=balance,
            balance_as_of=balance_as_of,
            currency=currency,
            institution_name=institution_name,
        ).dict(exclude_none=True)
        response_data = self._make_request(
            method=self.Methods.PUT,
            url_path=[APIConfig.LUNCHMONEY_ASSETS, asset_id],
            payload=payload,
        )
        asset = AssetsObject(**response_data)
        return asset

    def insert_asset(
        self,
        type_name: str,
        name: Optional[str] = None,
        subtype_name: Optional[str] = None,
        display_name: Optional[str] = None,
        balance: float = 0.00,
        balance_as_of: Optional[datetime.datetime] = None,
        currency: Optional[str] = None,
        institution_name: Optional[str] = None,
        closed_on: Optional[datetime.date] = None,
        exclude_transactions: bool = False,
    ) -> AssetsObject:
        """
        Create a single (manually-managed) asset.

        Parameters
        ----------
        type_name: Optional[str]
            Must be one of: cash, credit, investment, other, real estate, loan, vehicle,
            cryptocurrency, employee compensation
        name: Optional[str]
            Max 45 characters
        subtype_name: Optional[str]
            Max 25 characters
        display_name: Optional[str]
            Display name of the asset (as set by user)
        balance: float
            Numeric value of the current balance of the account. Do not include any
            special characters aside from a decimal point! Defaults to `0.00`
        balance_as_of: Optional[datetime.datetime]
            Has no effect if balance is not defined. If balance is defined, but
            balance_as_of is not supplied or is invalid, current timestamp will be used.
        currency: Optional[str]
            Three-letter lowercase currency in ISO 4217 format. The code sent must exist
            in our database. Defaults to user's primary currency.
        institution_name: Optional[str]
            Max 50 characters
        closed_on: Optional[datetime.date]
            The date this asset was closed
        exclude_transactions: bool
            If true, this asset will not show up as an option for assignment when
            creating transactions manually. Defaults to False

        Returns
        -------
        AssetsObject
        """
        payload = _AssetsParamsPost(
            type_name=type_name,
            subtype_name=subtype_name,
            name=name,
            display_name=display_name,
            balance=balance,
            balance_as_of=balance_as_of,
            currency=currency,
            institution_name=institution_name,
            closed_on=closed_on,
            exclude_transactions=exclude_transactions,
        ).dict(exclude_none=True)
        response_data = self._make_request(
            method=self.Methods.POST,
            url_path=[APIConfig.LUNCHMONEY_ASSETS],
            payload=payload,
        )
        asset = AssetsObject(**response_data)
        return asset
